;--------------------------------------------------------------------------
;
;	Create a fragment list of a file. The fragmentlist consists of
;	tuples of fragmentsize and fragmentstart.  Each tuple is a 32-bit
;	integer. This is to support direct block IO to disk images without
;	the need to read additional sectors from the device.
;
;	size	the size of this fragments in sectors
;	start	the number of the first sector covered by this fragment
;
;	When we now need to read or write a specific sector of a file we need
;	to find the corresponding entry. For this we compare the sector we 
;	want with the size of the fragment. If it is lower then we found the
;	fragment. Else we subtract the size from the sector which can be looked
;	at as the offset into the next fragment. If the calculated value is
;	less then the size of the next fragment we are done. Else we continue
;	until the number of sector is negative (if a file is 4 sectors long
;	the sectors we can read are numbered 0,1,2,3. When we subtract the
;	legnths of all fragements from the sector to read the result is -1)
;
;	When we found a fragment we just need to add the (remaining) sector
;	offset to the start sector number of this fragment. The results is
;	then the absolute sector on the device.
;
;	Input:
;	Y	Pointer to file control block
;
;	Output:
;	Fragmentlist created
;
;	Registers:
;	r16, r17, r18, r19, r20, r24, r25, zl, zh
;
;	Completioncode:
;	CS	The file has more fragments than we can store in memory
;	CC	Created complete fragment list
;
;
;	2019-01-03	Use linked fragment list. That is we first allocate a
;			fragment descriptor buffer and link it to the previous
;			one (or in case if it is the first we link it to the
;			fragment descriptor queue head of the IO block). 
;			We support as many fragments as we have free memory.
;
;******	20220-05-10	New register usage, list of caller:
;******			attachcmd		ok
;******
;******
;******
;
BuildFragList:
	movw	zl:zh, yl:yh
	adiw	Z, fcb_fraglist		; Queue Head Address
;
BuildFragListNext:
	ldi	r24, low(Fr_Entry)	; Get a memory block for one fragment entry
	ldi	r25, high(Fr_Entry)
	call	malloc
	brcs	BuildFragListMemfail
	std	Z+0, r24		; Copy address to the previous queue head
	std	Z+1, r25
	movw	zh:zl, r25:r24		; New queue head
	std	Z+0, zero		; Current end of chain
	std	Z+1, zero
	lds	r20, sectperclst	; initialise packet with a fragment of at least
	std	Z+Fr_Size+0, r20	; one cluster
	std	Z+Fr_Size+1, zero
	std	Z+Fr_Size+2, zero
	std	Z+Fr_Size+3, zero	; initial size of fragment = sectors per cluster
	rcall	Cluster2Sector		; convert cluster to sector
	ldd	r16, Y+P_Sector+0	; And make this the 
	ldd	r17, Y+P_Sector+1
	ldd	r18, Y+P_Sector+2
	ldd	r19, Y+P_Sector+3
	std	Z+Fr_Start+0, r16
	std	Z+Fr_Start+1, r17
	std	Z+Fr_Start+2, r18
	std	Z+Fr_Start+3, r19	; start sector of this fragment.
BuildFragListLoop:
	ldd	r16, Y+P_Cluster+0	; Save current cluster in the P_Sector
	ldd	r17, Y+P_Cluster+1	; field of the data structure which
	ldd	r18, Y+P_Cluster+2	; is currently not used
	ldd	r19, Y+P_Cluster+3
	std	Y+P_Sector+0, r16
	std	Y+P_Sector+1, r17	;
	std	Y+P_Sector+2, r18	;
	std	Y+P_Sector+3, r19	;
	rcall	LinkedCluster		; Get linked cluster
	breq	BuildFragListDone	; No more clusters
	ldd	r16, Y+P_Sector+0	; Get previous cluster
	ldd	r17, Y+P_Sector+1	; 
	ldd	r18, Y+P_Sector+2	; 
	ldd	r19, Y+P_Sector+3	; 
	subi	r16, byte1(-1)		; Increment by one
	sbci	r17, byte2(-1)
	sbci	r18, byte3(-1)
	sbci	r19, byte4(-1)
	ldd	r20, Y+P_Cluster+0
	cp	r20, r16
	ldd	r20, Y+P_Cluster+1
	cpc	r20, r17
	ldd	r20, Y+P_Cluster+2
	cpc	r20, r18
	ldd	r20, Y+P_Cluster+3
	cpc	r20, r19
	brne	BuildFragListNext	; need new fragement entry
;
	ldd	r16, Z+Fr_Size+0		; next cluster is adjacent, i.e. it belongs
	ldd	r17, Z+Fr_Size+1		; to this fragment, so we just
	ldd	r18, Z+Fr_Size+2		; 
	ldd	r19, Z+Fr_Size+3		; 
	lds	r20, sectperclst
	add	r16, r20			; add sectors per cluster to this fragmentsize
	adc	r17, zero
	adc	r18, zero
	adc	r19, zero
	std	Z+Fr_Size+0, r16
	std	Z+Fr_Size+1, r17
	std	Z+Fr_Size+2, r18
	std	Z+Fr_Size+3, r19		; save
	rjmp	BuildFragListLoop	; check next clusters
;
BuildFragListMemfail:
	movw	zl:zh, yl:yh
	adiw	Z, fcb_fraglist		; Queue Head Address
	rcall	FreeList	
	sec
	ret
;
BuildFragListDone:
	clc
	ret 
