;--------------------------------------------------------------------------
;
;	Subroutine for LinkedCluster
;	Converts Cluster to sector offset in FAT adds fat1start and
;	checks if the sector is already in pfatbuffer. Else it reads
;	the sector to pfatbuffer. Then it calculates the offset into
;	the sector and reads the linked cluster and checks if the end
;	of the linked list has been reached.
;	Input:
;	Y		datastructure of file
;
;	Output:
;	P_Cluster updated with linked cluster
;
;	Completioncode:
;	EQ		end of file reached no more clusters
;	NE		new cluster stored int P_Cluster
;
;	Registers:
;	r16, r17
;
;******	20220-05-10	New register usage, list of caller:
;******			BuildFraglist		ok
;******			ReadDir			ok
;******			ReadFileByte		ok
;******			linkedclust		ok
;******
LinkedCluster:
	push	r4
	push	r5
	push	r6
	push	r7
	push	zl
	push	zh			; save
	ldd	r4, Y+P_Cluster+0	; Retrieve the Cluster
	ldd	r5, Y+P_Cluster+1	; 
	clr	r6
	clr	r7			; Assume FAT16
;
;	Calculate Sector in FAT of current cluster
;	FAT16: 256 entries per sector, just shift one byte right
;	Sector = Fat1Start + pcluster/256

	lds	r17, VolStatus
	sbrs	r17, FAT32
	rjmp	LinkedCluster16
;
;	FAT32: 128 entries per sector, first shift one bit
;	to the left and then shift one byte to the right
;	Sector = Fat1Start + pcluster/128

	ldd	r6, Y+P_Cluster+2	; For FAT32 it is 4 bytes
	ldd	r7, Y+P_Cluster+3	;

	add	r4, r4			; In case of FAT32 we need to
	adc	r5, r5			; shift the Cluster by 2 in order
	adc	r6, r6			; to convert cluster in Sector as
	adc	r7, r7			; a FAT entry has 4 bytes
LinkedCluster16:
	push	yl
	push	yh			; Save current data structure pointer
	ldi	yl, low(pfatcluster)	; All FAT operations use a dedicated
	ldi	yh, high(pfatcluster)	; datastructure
;
;	Shifting the cluster one byte to the right is the same as
;	omitting the lowst 8 bits, which is later needed as the index
;	into the sector. We already took into account above that FAT32
;	entries are 4 bytes by putting cluster*2 into R4,R5,R6,R7
;
	lds	r17, fat1start+0
	add	r5, r17
	lds	r17, fat1start+1
	adc	r6, r17
	lds	r17, fat1start+2
	adc	r7, r17
	lds	r17, fat1start+3
	adc	r17, zero		; Sector of FAT sector we need
	
	ldd	r16, Y+P_Sector+0	; Check if we already have it
	cp	r16, r5
	brne	LinkedReadSect
	ldd	r16, Y+P_Sector+1
	cp	r16, r6
	brne	LinkedReadSect
	ldd	r16, Y+P_Sector+2
	cp	r16, r7
	brne	LinkedReadSect
	ldd	r16, Y+P_Sector+3
	cp	r16, r17
	breq	LinkedSameSect		; We already have it

LinkedReadSect:
	std	Y+P_Sector+0, r5
	std	Y+P_Sector+1, r6
	std	Y+P_Sector+2, r7
	std	Y+P_Sector+3, r17	; Save Sector to read
	call	SD_CARD_READ
LinkedDead:
	brcs	LinkedDead

LinkedSameSect:
	ldd	zl, Y+P_Address+0
	ldd	zh, Y+P_Address+1
	add	zl, r4			; r4 is half the offset into the FAT Sector
	adc	zh, zero		; so we have to add it twice
	add	zl, r4
	adc	zh, zero
	pop	yh
	pop	yl			; Restore calling data structure pointer

	lds	r17, VolStatus
	sbrc	r17, FAT32
	rjmp	LinkedCluster32
;
;	At the same time we copy the new cluster we also check the end of cluster
;	chain. A FAT16 entry of 0xFFF8..0xFFFF defines end of chain.
;
	std	Y+P_Cluster+2, zero
	std	Y+P_Cluster+3, zero	; Cluster is only a 16-bit value
	ld	r16, Z+
	ld	r17, Z+
	std	Y+P_Cluster+0, r16
	std	Y+P_Cluster+1, r17
	andi	r16, 0xF8
	cpi	r16, 0xF8
	brne	LinkedExit
	cpi	r17, 0xFF
	rjmp	LinkedExit
;
;
;
LinkedCluster32:
;
;	At the same time we copy the new cluster we also check the end of cluster
;	chain. A FAT32 entry of 0x0FFFFFF8..0x0FFFFFFF defines end of chain.
;
	ld	r16, Z+
	ld	r5, Z+
	ld	r6, Z+
	ld	r17, Z+
	std	Y+P_Cluster+0, r16
	std	Y+P_Cluster+1, r5
	std	Y+P_Cluster+2, r6
	std	Y+P_Cluster+3, r17

	andi	r16, 0xF8
	cpi	r16, 0xF8
	brne	LinkedExit
	ldi	r16, 0xff
	cp	r5, r16
	brne	LinkedExit
	cp	r6, r16
	brne	LinkedExit
	cpi	r17, 0x0F
LinkedExit:
	pop	zh
	pop	zl
	pop	r7
	pop	r6
	pop	r5
	pop	r4
	ret				; EQ if current cluster was last, else NE