;--------------------------------------------------------------------------
;
;	Translate logical block to physical sector using the fragment list
;	of the file entry
;
;	2019-07-26	Input changed from P_Sector to P_Cluster so we can keep the
;			logical block number in P_Cluster. As this is only used for
;			block IO files P_Cluster is not used as we have a fragment
;			list to access the sectors of the file.
;
;	Input:
;	Y		Pointer to data structure for file IO
;	Y+P_Cluster	Logical Block Number
;
;	Output;
;	Y+P_Sector	Physical Sector Number
;
;	Condition:
;	CC		Translation was successfull
;	CS		Translation failed
;
;	Registers:
;	r16, r17, r18, r19, r20
;
;	2019-01-03	Use linked fragment list
;	2019-07-24	Input is now expected in P_Cluster
;	2019-07-27	Save and restore registers
;
;******	20220-05-13	New register usage, list of caller:
;******			cmddumpbl		ok
;******			readblock		ok
;******			rlv12_autobootfile	ok
;******			rlv12_rwsetup		ok
;******			rlv12_rwnextsector	ok
Logical2Physical:
	push	zl
	push	zh
	ldd	zl, Y+fcb_fraglist+0
	ldd	zh, Y+fcb_fraglist+1	; Get first fragment descriptor
	ldd	r16, Y+P_Cluster+0	; Get logical block number
	ldd	r17, Y+P_Cluster+1
	ldd	r18, Y+P_Cluster+2
	ldd	r19, Y+P_Cluster+3

Logical2Loop:
	ldd	r20, Z+Fr_Size+0	; get size of current fragment list
	cp	r16, r20
	ldd	r20, Z+Fr_Size+1
	cpc	r17, r20
	ldd	r20, Z+Fr_Size+2
	cpc	r18, r20
	ldd	r20, Z+Fr_Size+3
	cpc	r19, r20
	brlo	Logical2Found		; logical block < size -> found
	
	ldd	r20, Z+Fr_Size+0	; Each entry in the fragment list
	sub	r16, r20		; has the size of the fragment in 
	ldd	r20, Z+Fr_Size+1	; sectors and the physical sector number
	sbc	r17, r20		; of the first sector of this fragment
	ldd	r20, Z+Fr_Size+2	; if we logical block is not in this
	sbc	r18, r20		; fragment we subtract the size of the
	ldd	r20, Z+Fr_Size+3	; fragment from the logical block and
	sbc	r19, r20		; check if it is within the next fragment.
	ldd	r20, Z+0		; Get next fragment descriptor
	ldd	zh, Z+1
	mov	zl, r20
	or	r20, zh		; check for valid address
	brne	Logical2Loop
	sec				; requested sector is beyound size
	rjmp	Logical2Exit
	ret
;
Logical2Found:
	ldd	r20, Z+Fr_Start+0	; Add physical sector number of first
	add	r16, r20		; sector in this fragment
	ldd	r20, Z+Fr_Start+1
	adc	r17, r20
	ldd	r20, Z+Fr_Start+2
	adc	r18, r20
	ldd	r20, Z+Fr_Start+3
	adc	r19, r20

	std	Y+P_Sector+0, r16	; return physical block number
	std	Y+P_Sector+1, r17
	std	Y+P_Sector+2, r18
	std	Y+P_Sector+3, r19
	clc
Logical2Exit:
	pop	zh
	pop	zl
	ret 
