;--------------------------------------------------------------------------
;
;	Read Directory Entry and returns a pointer either to the next active
;	or free directory entry if one exists. If no free directory entry
;	exists the pointer is set to zero. To read a directory first call
;	OpenDir with the P_Cluster set to the first cluster of the directory
;	file (or 0 in case you want to read the root directory). OpenDir and
;	ReadDir automatically handle the special case of FAT16 root directory.
;	
;	Input:
;	Y		Pointer to the data structure for directory
;	P_LongFileN	must be set to a buffer that will hold the
;			long filename
;	Output:
;	P_Entry		copy of the primary directory entry
;
;	Completioncode:
;	CC	Active Entry found
;	CS	Free or no entry found 
;
;	Registers:
;	r17
;
ReadDir:
	push	xl
	push	xh
	push	zl
	push	zh
	ldd	r17, Y+P_Flag
	cbr	r17, 1<<D_Long
	std	Y+P_Flag, r17		; No Long File Name so far
ReadDirEntry:
	ldd	r17, Y+P_Count		; Get number of entries already processed
	inc	r17			; Another entry processed
	std	Y+P_Count, r17
	cpi	r17, (512/32)		; 
	brne	ReadDirThis
ReadNxtDir:				; We require the next sector of the direcotry
	ldd	r17, Y+P_NumSect	; Decrement sectors to read
	dec	r17			; done in directory
	std	Y+P_NumSect, r17	; 
	brne	ReadNxtDirSect		; Just read the next sector (P_Sector++)
	ldd	r17, Y+P_Flag		; Check directory type
	sbrs	r17, D_Linked		; Linked list of clusters (normal file)
	rjmp	ReadNxtDirEnd		; In case of FAT16 root directory no more sectors
;
;	Either a normal directory or the FAT32 root directory is being processed which
;	are built like any file as a list of linked clusters. We reach here when we have
;	processed all sectors in the current cluster so we need to find the next cluster.
;
	rcall	LinkedCluster		; Follow Cluster List
	breq	ReadNxtDirEnd
	rcall	Cluster2Sector
	lds	r17, sectperclst	; Re-initialise number of 
	std	Y+P_NumSect, r17	; sectors in cluster.
	rjmp	ReadNxtReadSect
;
;	Read next directory sector of the current cluster or FAT16 root directory
;
ReadNxtDirSect:
	ldd	r17, Y+P_Sector+0	; Increment the sector number and read
;	add	r17, one		; sector
	subi	r17, byte1(-1)
	std	Y+P_Sector+0, r17
	ldd	r17, Y+P_Sector+1
;	adc	r17, zero
	sbci	r17, byte2(-1)
	std	Y+P_Sector+1, r17
	ldd	r17, Y+P_Sector+2
;	adc	r17, zero
	sbci	r17, byte3(-1)
	std	Y+P_Sector+2, r17
	ldd	r17, Y+P_Sector+3
;	adc	r17, zero
	sbci	r17, byte4(-1)
	std	Y+P_Sector+3, r17
ReadNxtReadSect:
	ldi	r17, 0xff
	std	Y+P_Count, r17		; No directory entries in sector processed
	ldi	r17, low(pdirbuffer)	; Reset the buffer pointer
	std	Y+P_Address+0, r17
	ldi	r17, high(pdirbuffer)
	std	Y+P_Address+1, r17
	ldi	r17, led_time
	sts	led_oneshot, r17
	sbi	b_LED
	call	SD_CARD_READ
	brcc	ReadDirEntry		; Re-enter
ReadNxtDirEnd:
	sec
	std	Y+P_DirPointer+0, zero	
	std	Y+P_DirPointer+1, zero	
	pop	zh
	pop	zl
	pop	xh
	pop	xl
	ret

ReadDirThis:
	ldd	zl, Y+P_Address+0	; Get address of next directory entry
	ldd	zh, Y+P_Address+1
	ldd	r17, Z+D_Name		; Get first character of Name
	tst	r17
	breq	ReadDirEnd		; if it is 0x00 end of directory
	cpi	r17, 0xe5
	breq	ReadDirNext		; This is a deleted entry, skip it
	ldd	r17, Z+D_Attr		; is it part of a long filename
	cpi	r17, A_Long
	breq	ReadDirLong
	std	Y+P_DirPointer+0, zl
	std	Y+P_DirPointer+1, zh
	adiw	Z, 32			; Next Entry Address
	std	Y+P_Address+0, zl	; Save for next entry to this routine
	std	Y+P_Address+1, zh
	pop	zh
	pop	zl
	pop	xh
	pop	xl
	clc
	ret
ReadDirLong:				; Process a long filename entry
	ldd	r17, Y+P_Flag
	sbrc	r17, D_Long
	rjmp	ReadDirLongCont		; If set then continue with long filename
	sbr	r17, 1<<D_Long
	std	Y+P_Flag, r17
	ldd	r17, Z+D_Name
	sbrs	r17, 6			; We expect this to be the last entry
	rjmp	ReadDirNext		; Something is wrong with directory, skip it
	ldd	xl, Y+P_LongFileN+0
	ldd	xh, Y+P_LongFileN+1
	clr	r17
ReadDirLongClr:
	st	X+, zero
	inc	r17
	brne	ReadDirLongClr		; Initialise long filename buffer with 0x00
	rjmp	ReadDirCopy
ReadDirLongCont:
	ldd	r17, Z+D_Name
	sbrc	r17, 6			; 
	rjmp	ReadDirNext		; If "Last" flag set this is an error, skip it
ReadDirCopy:
	rcall	ReadDirCopy13
ReadDirNext:
	adiw	Z, 32			; Next Entry Address
	std	Y+P_Address+0, zl	; Save for next entry to this routine
	std	Y+P_Address+1, zh
	rjmp	ReadDirEntry
;
ReadDirEnd:
	std	Y+P_Address+0, zl	; Save for next entry to this routine
	std	Y+P_Address+1, zh
	pop	zh
	pop	zl
	pop	xh
	pop	xl
	sec
	ret
;
;	Copy the characters from a long filename directory entry to the
;	long filename buffer. Long filenames are split into directory
;	entires each with up to 13 double-byte characters. The first
;	byte if the filename has the index of the part. So we just
;	need to multiply the index minus one with 13 and copy the 13
;	characters. Note we assume it is all ASCII, that is the
;	highbyte of each double-byte character is 0. We do not even check
;	as this would also require that we are able to support this but
;	we are not for the moment. 
;
ReadDirCopy13:
	ldi	r17, 13		; Each entry has up to 13 characters
	mov	r0, r17		; save constant
	ldd	r17, Z+D_Name	; get first byte of filename which 
	andi	r17, 0x3F	; contains the index, isolate the index
	dec	r17		; offset=(entrynum-1)*13
	mul	r0, r17		; Calcualte offset of this entry into buffer

	ldd	xl, Y+P_LongFileN+0
	ldd	xh, Y+P_LongFileN+1
	add	xl, r0
	adc	xh, zero	; we assume that filenames are less than 255 characters long

	ldd	r17, Z+D_Name+1	; copy the 13 locations, note that the
	st	X+, r17		; string in the extension is 0x0000 terminated
	ldd	r17, Z+D_Name+3	; we assume only ASCII double characeters
	st	X+, r17
	ldd	r17, Z+D_Name+5
	st	X+, r17
	ldd	r17, Z+D_Name+7
	st	X+, r17
	ldd	r17, Z+D_Name+9
	st	X+, r17
	ldd	r17, Z+D_Name+14
	st	X+, r17
	ldd	r17, Z+D_Name+16
	st	X+, r17
	ldd	r17, Z+D_Name+18
	st	X+, r17
	ldd	r17, Z+D_Name+20
	st	X+, r17
	ldd	r17, Z+D_Name+22
	st	X+, r17
	ldd	r17, Z+D_Name+24
	st	X+, r17
	ldd	r17, Z+D_Name+28
	st	X+, r17
	ldd	r17, Z+D_Name+30
	st	X+, r17
	ret
 
