;--------------------------------------------------------------------------
;
;	Controller Functions Log Entries have the following format. 
;	A log entry is only valid if the first byte is not zero and the
;	MSB of the first byte is cleared. I.e. only values between 1..127
;	mark a valid log entry. 
;
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	| 0 | 0 |A21|A20|A19|A18|A17|A16| 0 | IE|DS1|DS0|FC2|FC1|FC0|DRY|
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	|A15|A14|A13|A12|A11|A10| A9| A8| A7| A6| A5| A4| A3| A2| A1| A0|
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	|                              DAR                              |
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	|                              MPR                              |
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;
;	General Log Entries have the following header. As I did not 
;	implement the maintenance function (FC=0) this cannot conflict  
;	with the controller functions log entries.
;
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	|           Log Code            | 0 | 0 | 0 | 0 | 0 | 0 | 0 | 1 | 
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	...
;
;	Log Code
;	0	Write Device Register
;	1	Read Device Register
;	2	Physical Block number
;	3	Interrupt
;	4	Read ROM
;	5	Write ROM
;	6	Autoboot
;	7	Bus INIT
;	8	Trace
;
;
;	The format for code 2 and 6 is
;
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	|           Log Code            | 0 | 0 | 0 | 0 | 0 | 0 | 0 | 1 | 
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	|                                                               |
;	+--  int32 first pysical block number on SD-Card of request  ---+
;	|                                                               |
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	|            number of blocks transferred in request            |
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;
;	The format for the other codes is
;
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	|           Log Code            | 0 | 0 | 0 | 0 | 0 | 0 | 0 | 1 | 
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	|                            Bus Data (*)                       |
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	|       TCNT2 Tick Value        |   Bus Address if applicable   |
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	|                             unused                            |
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;
;	(*) BINIT creates 2 entries, one for asserting and one for 
;	deasserting BINIT, instead of Bus Data this field has the
;	the BINIT status
;
logprint:
	push	xl
	push	xh
	push	yl
	push	yh
	push	zl
	push	zh
	in	r19, GPIOR0		; Keep a copy of logging status
	out	GPIOR0, zero		; Stopp Logging
	ldi	r17, high(log_begin)
	tst	r17
	breq	logprint130		; No lollipop
	call	print
	.db	"Printing Initial Logging ", CR, LF, 0
	ldi	yl, low(log_buffer)
	ldi	yh, high(log_buffer)
	ldi	xl, low(log_begin)
	ldi	xh, high(log_begin)
logprint110:
	ldd	r17, Y+0
	tst	r17
	breq	logprint120
	brmi	logprint120
	rcall	logentry
logprint120:
	adiw	yh:yl, 8
	sbiw	xh:xl, 8
	brpl	logprint110
	rcall	logask
	brcs	logprint040
logprint130:
	call	print
	.db	CR, LF, "Printing Circular Logging", CR, LF, 0

	lds	yl, log_pointer+0
	lds	yh, log_pointer+1
	ldi	xl, low(log_size)
	ldi	xh, high(log_size)
logprint010:
	ldd	r17, Y+0
	tst	r17
	breq	logprint020		; Must not be zero
	brmi	logprint020		; And MSB must be cleared
	rcall	logentry
logprint020:
	adiw	yh:yl, 8
	sbrc	yh, log_overflow	;;; 
	subi	yh, high(log_size)	;;; Go to start of circular buffer
	sbiw	xh:xl, 8
	brpl	logprint010
	
logprint040:
	ldi	yl, low(log_buffer)
	ldi	yh, high(log_buffer)
	sts	log_pointer+0, yl
	sts	log_pointer+1, yh
	ldi	xl, low(log_buffersize)
	ldi	xh, high(log_buffersize)
	ldi	r18, 0xff
logini:
	st	Y+, r18
	sbiw	xh:xl, 1
	brne	logini
	out	GPIOR0, r19		; Resume Logging
	pop	zh
	pop	zl
	pop	yh
	pop	yl
	pop	xh
	pop	xl
	ret
;
;logentrymsg:
;	call	print
;	.db		"Message buffer:", LF, CR, 0
logentry:
;
;	Function Logg Entries all have the same header information, therefore
;	we set first the parameter block for the print routine for all function
;	codes
;
;	pprint+0:	The disk select bits converted to character 0..3
;	pprint+1:	-not used-
;	pprint+2:	Address Bits A0..7
;	pprint+3:	Address Bits A8..15
;	pprint+6:	Multi Purpose Register Value low
;	pprint+7:	Multi Purpose Register Value high
;
;
	ldd	zl, Y+0			; 
	mov	r17, zl			; 
	swap	r17			; Isolate drive select bits
	andi	r17, driveselect	; 
	ori	r17, '0'		; Translate to ASCII number
	sts	pprint+0, r17		; Volume as Character
	ldd	r17, Y+2		; Copy Part of Log Entry 
	sts	pprint+2, r17		;
	ldd	r17, Y+3		;
	sts	pprint+3, r17		;
	ldd	r17, Y+6		;
	sts	pprint+6, r17		;
	ldd	r17, Y+7		;
	sts	pprint+7, r17		;
	lsr	zl			; Shift function code to bits0..2
	andi	zl, 0x07		; Isolate function code
	clr	zh
	subi	zl, low(-logtable)	; Prepare for select (ijmp)
	sbci	zh, high(-logtable)
	ijmp				; 
	;;;;
logtable:
	rjmp 	logprint_maint
	rjmp 	logprint_writechk
	rjmp 	logprint_getstatus
	rjmp 	logprint_seek
	rjmp 	logprint_readheader
	rjmp 	logprint_writedata
	rjmp 	logprint_readdata
	rjmp 	logprint_readnocheck


logask:
	call	print
	.db	"Continue [Y]", 0, 0
	call	serin
	cpi	r24, 'y'
	breq	logask010
	cpi	r24, 'Y'
	breq	logask010
	cpi	r24, CR
	breq	logask010
	cpi	r24, 'n'
	breq	logask020
	cpi	r24, 'N'
	breq	logask020
	call	seroutcrlf
	rjmp	logask
logask010:
	clc
	ret
logask020:
	sec
	ret
;--------------------------------------------------------------------------
;
;	Each rlv12 function stores an 8-byte entry in the circular logging
;	buffer. If an 8-byte entry starts with 0xFF or 0x00 it is considered
;	as invalid and skipped. 
;
;	"show logging" analyzes the bytes and converts them to a readable format
;	and then initialises the ring buffer with 0xff. It initialises after the
;	the "show logging" as then the buffer is preserved during a reset.
;
logprint_maint:
;
;	Maintenance has been reused for register logging
;
	ldd	r17, Y+4		; Bus Address
	sts	pprint+0, r17
	ldd	r17, Y+5		; ticks
	sts	pprint+1, r17
	ldd	zl, Y+1			; Get subcode
	cpi	zl, log_undef		; must be set to first invalid number
	brsh	logprint_unkown
	clr	zh
	subi	zl, low(-logtablem)
	sbci	zh, high(-logtablem)
	ijmp
logtablem:
	rjmp	logprint_regread
	rjmp	logprint_regwrite
	rjmp	logprint_pbn
	rjmp	logprint_iack
	rjmp	logprint_romread
	rjmp	logprint_romwrite
	rjmp	logprint_autoboot
	rjmp	logprint_init
	rjmp	logprint_trace
	
logprint_unkown:
	call	print					
	.db	"unknown (", 0x80,")", 0	; anything else is not defined
	rjmp	logprint_regprint
	
logprint_regread:
	call	print
	.db	"Read  (", 0x80, ",t:", 0x81, ")", 0
	rjmp	logprint_regprint
logprint_regwrite:
	call	print
	.db	"Write (", 0x80, ",t:", 0x81, ")", 0
	rjmp	logprint_regprint
;
;
;
logprint_iack:
	call	print
	.db	"IACK     (t:", 0x81, ") VEC:", 0xa2, CR, LF, 0, 0
	ret

logprint_init:
	ldd	r17, Y+2		; Init FLAG
	ori	r17, '0'
	sts	pprint+2, r17
	call	print
	.db	"INIT     (t:", 0x81, ") i_INIT:", 0x92, CR, LF, 0
	ret
;
;	Autoboot
;
;
logprint_autoboot:
	ldd	r17, Y+2		; Physical Block Number
	sts	pprint+0, r17
	ldd	r17, Y+3
	sts	pprint+1, r17
	ldd	r17, Y+4
	sts	pprint+2, r17
	ldd	r17, Y+5
	sts	pprint+3, r17
	ldd	r17, Y+6		; csrl
	sts	pprint+4, r17
	ldd	r17, Y+7
	sts	pprint+5, r17
	call	print
	.db	"Autoboot  P:0x", 0x83, 0x82, 0x81, 0x80, " N:0x", 0x85, 0x84, CR, LF, 0
	ret
;
;	Start Physical Block + Blockcount
;
;
logprint_pbn:
	ldd	r17, Y+2		; Physical Block Number
	sts	pprint+0, r17
	ldd	r17, Y+3
	sts	pprint+1, r17
	ldd	r17, Y+4
	sts	pprint+2, r17
	ldd	r17, Y+5
	sts	pprint+3, r17
	ldd	r17, Y+6		; Number of blocks transferred from SD-Card
	sts	pprint+4, r17
	ldd	r17, Y+7
	sts	pprint+5, r17
	call	print
	.db	"          P:0x", 0x83, 0x82, 0x81, 0x80, " N:0x", 0x85, 0x84, CR, LF, 0
	ret
	
logprint_romread:
	call	print
	.db	"ROM Read  (", 0x80, ",t:", 0x81, ") ", 0xa2, CR, LF, 0
	ret
logprint_romwrite:
	call	print
	.db	"ROM Write (", 0x80, ",t:", 0x81, ") ", 0xa2, CR, LF, 0
	ret
;
;
;
logprint_regprint:
	ldd	zl, Y+4
	andi	zl, 0x07
	add	zl, zl
	add	zl, zl
	clr	zh
	subi	zl, low(-2*logprint_names)
	sbci	zh, high(-2*logprint_names)
	lpm	r17, Z+
	sts	pprint+6, r17
	lpm	r17, Z+
	sts	pprint+7, r17
	lpm	r17, Z+
	sts	pprint+8, r17
	call	print
	.db	" ", 0x96, 0x97,0x98, ":", 0xa2, CR, LF, 0, 0
	ret
;
;
;
logprint_names:
	.db	"CSR BAR DAR MPR BAE 012 014 016 "

	
logprint_trace:
	ldd	r17, Y+2		; Physical Block Number
	sts	pprint+0, r17
	ldd	r17, Y+3
	sts	pprint+1, r17
	ldd	r17, Y+4
	sts	pprint+2, r17
	ldd	r17, Y+5
	sts	pprint+3, r17
	ldd	r17, Y+6		; Number of blocks transferred from SD-Card
	sts	pprint+4, r17
	ldd	r17, Y+7
	sts	pprint+5, r17
	call	print
	.db	"Trace: 0x", 0x80, ", 0x", 0x81, ", 0x", 0x83, 0x82, CR, LF, 0
	ret

;--------------------------------------------------------------------------
;
;	Print Function Code Logging
;
;
logprint_readnocheck:
	ldd	r17, Y+0
	sts	pprint+0, r17
	ldd	r17, Y+1
	sts	pprint+1, r17
	ldd	r17, Y+2
	sts	pprint+2, r17
	ldd	r17, Y+3
	sts	pprint+3, r17
	ldd	r17, Y+4
	sts	pprint+4, r17
	ldd	r17, Y+5
	sts	pprint+5, r17
	ldd	r17, Y+6
	sts	pprint+6, r17
	ldd	r17, Y+7
	sts	pprint+7, r17

	sts	pprint+8, yl
	sts	pprint+9, yh

	call	print
	.db	"no decoder 0x", 0x89, 0x88, ":"
	.db	0x80, " ", 0x81, " ", 0x82, " ", 0x83, " "
	.db	0x84, " ", 0x85, " ", 0x86, " ", 0x87, " "
	.db	CR, LF, 0, 0
	ret
;
;
;
logprint_readheader:
	call	print
	.db	"RDHD: V:", 0x90, " D:", 0xa6, CR, LF, 0
	ret
;
;
;
logprint_getstatus:
	call	print
	.db	"GETS: V:", 0x90, " S:", 0xa6, CR, LF, 0
	ret
;
;
;
logprint_seek:
	ldd	r16, Y+4
	ldd	r17, Y+5
	ldi	r18, '-'			; -> border
	sbrc	r16, 2				;
	ldi	r18, '+'			; -> spindle
	sts	pprint+8, r18
	ldi	r18, '0'
	sbrc	r16, 4				; -> Set Header
	ldi	r18, '1'
	sts	pprint+9, r18
	call	print
	.db	"SEEK: V:", 0x90, " C:", 0xa2, " O:", 0x98, 0
;
;	Convert Offset to ASCII Octal
;
	ldi	r18, 3
logprint_seek010:
	rol	r16
	rol	r17
	rol	r24
	rol	r16
	rol	r17
	rol	r24
	rol	r16
	rol	r17
	rol	r24
	andi	r24, 0x07
	ori	r24, '0'
	call	serout
	dec	r18
	brne	logprint_seek010

	call	print
	.db	"->", 0xa6, " H:", 0x99, CR, LF, 0
	ret
;
;	Write Check, Write and Read have the same format, first we print
;	the function code and jump to the common code
;
logprint_writechk:
	call	print
	.db	"WCHK", 0, 0
	rjmp	logprint_data
	
logprint_writedata:
	call	print
	.db	"WRIT", 0, 0
	rjmp	logprint_data

logprint_readdata:
	call	print
	.db	"READ", 0, 0
;
;	Prepare the remaining part of the parameter block for print
;
;	pprint+4:	Bus Address Extension
;	pprint+8:	Disk Address Register low
;	pprint+9:	Disk Address Register high
;	pprint+10:	Logical Block Number low
;	pprint+11:	Logical Block Number high
;
logprint_data:
	push	xl
	push	xh			; Must not change loop counter 
	ldd	r17, Y+1		; Bus Address Extension
	andi	r17, 0x3F		; Isolate bits
	sts	pprint+4, r17		; 

	ldd	r24, Y+4			; DAR
	ldd	r25, Y+5
	sts	pprint+8, r24
	sts	pprint+9, r25
;
;	Convert Disk address to logical block
;
	call	rlv12_dar2lbn

	sts	pprint+10, r24
	sts	pprint+11, r25

	call	print
	.db	": V:", 0x90, " A:", 0xb2, "(0x", 0x84, 0x83, 0x82, ") D:", 0xa8
	.db	"(L:0x", 0x8b, 0x8a, ") C:", 0xa6, 0, 0
;
;	Calculate and print the memory address after the DMA
;
	ldd	r16, Y+2		; Bus Address Register low
	ldd	r17, Y+3		; Bus Address Register high
	ldd	r18, Y+1		; Bus Address Extension Register
	andi	r18, 0x3F		; Isolate address bits
	
	ldd	xl, Y+6			; Multi Purpose Register low
	ldd	xh, Y+7			; Multi Purpose Register high
;
;	xh:xl has now the word count as 2's complement, so we need to
;	subtract it twice from the starting address to get the address
;	after the DMA block
;

	sub	r16, xl
	sbc	r17, xh
	sbci	r18, -1
	sub	r16, xl
	sbc	r17, xh
	sbci	r18, -1
	
	sts	pprint+0, r16	
	sts	pprint+1, r17	
	sts	pprint+2, r18	
	call	print
	.db	" E:", 0xb0, "(0x", 0x82, 0x81, 0x80, ")", LF, CR, 0
	pop	xh
	pop	xl
	ret

;==========================================================================
;
;	
;
logrom:
	lds	r16, tpflags
	sbrc	r16, tp__no
	rjmp	logromno
	sbi	GPIOR0, log__rom
	clc
	ret
logromno:
	cbi	GPIOR0, log__rom
	clc
	ret

logreg:
	lds	r16, tpflags
	sbrc	r16, tp__no
	rjmp	logregno
	sbi	GPIOR0, log__reg
	clc
	ret
logregno:
	cbi	GPIOR0, log__reg
	clc
	ret

logpbn:
	lds	r16, tpflags
	sbrc	r16, tp__no
	rjmp	logpbnno
	sbi	GPIOR0, log__pbn
	clc
	ret
logpbnno:
	cbi	GPIOR0, log__pbn

logiack:
	lds	r16, tpflags
	sbrc	r16, tp__no
	rjmp	logiackno
	sbi	GPIOR0, log__iack
	clc
	ret
logiackno:
	cbi	GPIOR0, log__iack
	clc
	ret

logunits:
	lds	r16, tpflags
	sbrc	r16, tp__no
	rjmp	logunitsno
	in	r16, GPIOR0
	ori	r16, log__units
	out	GPIOR0, r16
	clc
	ret
logunitsno:
	in	r16, GPIOR0
	andi	r16, ~log__units
	out	GPIOR0, r16
	clc
	ret
	

logunit:
	push	r18
	lds	r18, attunit
	ldi	r16, (1<<log__rl0)
logunit010:
	dec	r18
	brmi	logunit020
	lsl	r16
	rjmp	logunit010
logunit020:
	lds	r18, tpflags
	sbrc	r18, tp__no
	rjmp	logunitno
	in	r18, GPIOR0
	or	r18, r16
	out	GPIOR0, r18
	pop	r18
	clc
	ret
logunitno:
	com	r16
	in	r18, GPIOR0
	and	r18, r16
	out	GPIOR0, r18
	pop	r18
	clc
	ret
	
logstatus:
	lds	r16, tpflags
	sbrc	r16, tp__no
	rjmp	logallno
	in	r16, GPIOR0
;
	call	print
	.db	"Logging device register access .:", 0
	bst	r16, log__reg
	rcall	logstatusonoff
;
	call	print
	.db	"Logging boot ROM access ........:", 0
	bst	r16, log__rom
	rcall	logstatusonoff
;
	call	print
	.db	"Logging Interrupts .............:", 0
	bst	r16, log__iack
	rcall	logstatusonoff
;
	call	print
	.db	"Logging Physical Block Nbr. ....:", 0
	bst	r16, log__pbn
	rcall	logstatusonoff
;
	call	print
	.db	"Logging unit 0 .................:", 0
	bst	r16, log__rl0
	rcall	logstatusonoff
;
	call	print
	.db	"Logging unit 1 .................:", 0
	bst	r16, log__rl1
	rcall	logstatusonoff
;
	call	print
	.db	"Logging unit 2 .................:", 0
	bst	r16, log__rl2
	rcall	logstatusonoff
;
	call	print
	.db	"Logging unit 3 .................:", 0
	bst	r16, log__rl3
	rcall	logstatusonoff
;
	clc
	ret

logstatusonoff:
	brtc	logstatusonoff010
	call	print
	.db	" on", CR, LF, 0
	ret
logstatusonoff010:
	call	print
	.db	" off", CR, LF, 0, 0
	ret
	

logallno:
	out	GPIOR0, zero
	clc
	ret
