;--------------------------------------------------------------------------
;
;	Volume Primitives
;
;--------------------------------------------------------------------------
;
;	Mount Volume. Read MBR and VBR collect the information to the final
;	memory locations and calculate the startpositions of the various
;	regions. Set the flag for FAT32.
;
;	Only the first FAT partition will be mounted.
;
;	2018-02-15	MountVolume will now scan all the partitions in the
;			partition table. It will support various partition types
;			and map them to internal data structures. It also supports
;			now extended partitions. MountVolume will do the following:
;			-	It will scan the partition tables of the primary
;				and all extended partition.
;			-	If it encounters a partition of type LINUX (0x83) 
;				it will map each of them to one of the four emulated 
;				RL02 volumes
;			-	When it encounters a FAT partition (0x0b) it will 
;				map this partition to the FAT Volume control data
;				structure
;			Regardless of the number of partitions on the SD-Card it will
;			consider only the first 4 LINUX parititons and only the first
;			FAT16/32 partition it encounters. Additional partitions are
;			ignored.
;	2018-04-01	-	If it encounters a partition of type FAT12 (0x01) 
;				and the Q command has been called using the magic 
;				code 42 (i.e. Q hasbeen invoked with 42Q) then 
;				it will change the code from 0x01 (FAT12) to 
;				0x83 (Linux) and write the sector back to the SD-Card
;	2018-05-01	-	Partition Type FAT16B (0x06) is now handled correctly.
;	2018-12-27	New table drive detection of drive type as we need a more
;			flexible way for the RH70 to support all possible RM/RP 
;			drives.
;	2019-01-02	Partitions are kept in a list of linked partition control blocks
;	2019-06-29	New partition control block 
;			Also create a partition control block for FAT12 and correct size
;			A partition can have several stati
;			idle		a partition of correct size but incorrect 
;					partition type
;			active		a partition of correct size and correct 
;					partition type
;			assigned	a partition assigned to a drive letter, this
;					is either a FAT-16 or FAT-32 partition 
;	2019-07-03	StoreVolumeFAT just return in case we already have a FAT
;			Partition don't jump to Analyze030!!
;	2019-07-06	All partitions are now stored in partition control blocks
;	2019-07-27	Show less information during mount volume 
;			Add a command (fdisk) that shows the information on demand
;	2022-05-01	Clean-Up and make sure all registers are save because
;			MountVolume is called by sdcardinsert which itself is
;			called by serin, but serin must not change any register
;			except r24, the returned character.
;			
;	Input:
;	none
;
;	Output:
;	rc		set with return code
;
;	Conditions:
;	CC		Volume mounted successfully
;	CS		Mount volume failed
;
;	Note: MountVolume is part of the sdcardinsert process and hence
;	it must not change any register except r24, r25, because
;	sdcardinsert is called by the scheduler which is part of the
;	serin function and all callers of serin expect that nothing
;	except r24, r25 is changed.
;
;--------------------------------------------------------------------------
MountVolume:
	push	r4
	push	r16
	push	r17
	push	r18
	push	r19
	push	r20
	push	r22
	push	r23
	push	xl
	push	xh
	push	yl
	push	yh
	push	zl
	push	zh
	ldi	yl, low(pcluster)	; Prepare an IO control block
	ldi	yh, high(pcluster)	; 
	ldi	zl, low(pbuffer)	; set IO buffer address
	ldi	zh, high(pbuffer)	; 
	std	Y+P_Address+0, zl
	std	Y+P_Address+1, zh

	std	Y+P_Sector+0, zero	; Set sector to MBR
	std	Y+P_Sector+1, zero
	std	Y+P_Sector+2, zero
	std	Y+P_Sector+3, zero

	sts	pdirstartc+0, zero
	sts	pdirstartc+1, zero
	sts	pdirstartc+2, zero
	sts	pdirstartc+3, zero
	
	sts	Path, zero
		
	std	Y+P_Flag, zero		; No partition seen
	sts	VolStatus, zero		; No volume seen
	call	rlv12_init
;--------------------------------------------------------------------------
;
;	Read the current master record and analyze the partition table
;
AnalyzePartTable:
	call	SD_CARD_READ		; Read MBR
	brcc	AnalyzePartTable010
	ldi	r24, 0x81
	sec
	rjmp	AnalyzePartExit
AnalyzePartTable010:
	lds	r16, pbuffer+510
	cpi	r16, 0x55
	brne	AnalyzePartTable020
	lds	r16, pbuffer+511
	cpi	r16, 0xAA
	breq	AnalyzePart
AnalyzePartTable020:
	ldi	r24, 0x82
	sec
	rjmp	AnalyzePartExit
;
;	Type
;	0x01	FAT12	this partition type is used as candidate for RL01
;			RL02 volumes. There is an option to write back
;			the paritition table with partition code 0x83.
;
;	0x05	Ext.	Note that start sectors in the partition table of
;			an extended partition are relative to the start
;			of the extended partition
;	0x06	FAT16B	<......
;	0x07	exFAT	this partition type is used as candidate for MSCP volumes
;	0x0B	FAT32
;	0x83	Linux
;
;	There are always 4 partition entries in one partition table. There should
;	be only one extended partition per table. Normally it is but it has not to
;	be the last entry.
;
AnalyzePart:
	ldd	zl, Y+P_Address+0
	ldd	zh, Y+P_Address+1	
	subi	zl, low(-M_PartTable)	
	sbci	zh, high(-M_PartTable)	
	ldi	r18, 4
	mov	r4, r18			; Use a preserved register
AnalyzePartEntry:
	ldd	r16, Z+M_PartType
	cpi	r16, 0x01
	brne	AnalyzePart010
	ldi	r24, (1<<pcb__fat12)
	rcall	QueuePartImage
	rjmp	AnalyzePart100
	
AnalyzePart010:
;
;	Check for extended partition. Note that the extended partition does not
;	have to be the last in the partition table, but it should be the only
;	one. Therefore we just remember that we have found one and store the
;	start sector in P_Cluster. If there should be multiple extended partitions
;	only the first one is considered.
;
;	2018-03-31	Extended Partitions may have an extended partition however
;			the partition offset is the offset from the first extended
;			partitiond
;	2018-12-26	The offset of the primary extended partition is now stored
;			in pextended and the Sector offset of the next extended partition
;			is stored in the control block at P_Cluster
;
	cpi	r16, 0x05		; Is this an extended partition
	brne	AnalyzePart020
	rcall	AnalyzeExtended
	rjmp	AnalyzePart100
AnalyzePart020:
	cpi	r16, 0x83		; Linux
	brne	AnalyzePart030
	ldi	r24, 0
	rcall	QueuePartImage
	rjmp	AnalyzePart100
AnalyzePart030:
	cpi	r16, 0x06		; FAT-16B
	breq	AnalyzePart040
	cpi	r16, 0x0b		; FAT-32
	brne	AnalyzePart050
AnalyzePart040:
	ldi	r24, (1<<pcb__fat)
	rcall	QueuePartImage
	rjmp	AnalyzePart100
AnalyzePart050:
	;
AnalyzePart100:
	subi	zl, low(-M_PartEntry)	; Point to next entry in partition table
	sbci	zh, high(-M_PartEntry)
	dec	r4			; Another to go?
;
	brne	AnalyzePartEntry
	ldd	r16, Y+P_Flag
	sbrs	r16, Part_Next		; Do we have a next extended partition?
	rjmp	AnalyzePartDone		; no, then we are done
	cbr	r16, (1<<Part_Next)	; Reset Extended Partition Flag
	std	Y+P_Flag, r16		; else we would loop forever!
;
;	If there was an extended partition the start sector is now stored in P_Cluster
;	and we need to analyze the partition entries in the first sector of the extended
;	partition, which typically just holds one normal partition and a next extended 
;	partition entry.
;
	ldd	r16, Y+P_Cluster+0
	std	Y+P_Sector+0, r16		
	ldd	r16, Y+P_Cluster+1
	std	Y+P_Sector+1, r16
	ldd	r16, Y+P_Cluster+2
	std	Y+P_Sector+2, r16
	ldd	r16, Y+P_Cluster+3
	std	Y+P_Sector+3, r16
	rjmp	AnalyzePartTable


;--------------------------------------------------------------------------
;
;	The partition tables have now all been analyzed and if a partition
;	has been found that matches a drive then a partition control block
;	has been queued to the pcbqueue. Now we will attach the partitions
;	to the units
;
AnalyzePartDone:
	rcall	MountVolumeFAT
	call	checkvolume		; Do we have a FAT volume
	brcs	AnalyzePartNoVol	; Nope

	ldi	yl, low(pdircluster)	; Initialse read directory structure
	ldi	yh, high(pdircluster)

	ldi	r16, low(pdirbuffer)
	ldi	r17, high(pdirbuffer)
	std	Y+P_Address+0, r16
	std	Y+P_Address+1, r17

	ldi	r16, low(plfilename)
	ldi	r17, high(plfilename)
	std	Y+P_LongFileN+0, r16
	std	Y+P_LongFileN+1, r17

	std	Y+P_StartClust+0, zero	; Start with PWD set to root
	std	Y+P_StartClust+1, zero
	std	Y+P_StartClust+2, zero
	std	Y+P_StartClust+3, zero
	sts	Path, zero					;

	call	readinit
	cpi	r24, IO_SUCCESS
	brne	AnalyzePartNoVol
	rjmp	AnalyzePartSuccess

AnalyzePartNoVol:
	ldi	r18, units		; Number of units the controller supports
	ldi	xl, low(unittable)	; The unittable
	ldi	xh, high(unittable)
	ldi	yl, low(pcbqueue)
	ldi	yh, high(pcbqueue)
AnalyzePartNext:
	movw	zh:zl, yh:yl		; copy queue head pointer
	ldd	yl, Z+0			; get address of next pcb
	ldd	yh, Z+1
	sbiw	yh:yl, 0		; found another PCB
	brne	AnalyzePartNext010	; yes
	rjmp	AnalyzePartSuccess	; end of pcb list reached
AnalyzePartNext010:
	ldd	r16, Y+pcb_status
	sbrc	r16, pcb__fat12
	rjmp	AnalyzePartNext
	sbrc	r16, pcb__fat
	rjmp	AnalyzePartNext
;
;	Y is a valid partition control block
;
	sts	pprint+0, xl		; unit table address
	sts	pprint+1, xh
	sts	pprint+2, yl		; pcb address
	sts	pprint+3, yh
	ldd	zl, Y+pcb_drvtab+0	; Get drive table entry
	ldd	zh, Y+pcb_drvtab+1	;
	adiw	zh:zl, drv_typ		; Get drive type
	lpm	r16, Z
	sts	pprint+4, r16
	movw	Z, X
	std	Z+ucb_type, r16
	ldi	r16, (1<<ucb__rdy) | (1<<ucb__part)	; Mark Ready and attached to partition
	std	Z+ucb_status, r16
	sts	pprint+5, r16
	ldd	r16, Y+pcb_start+0
	sts	pprint+6, r16
	ldd	r16, Y+pcb_start+1
	sts	pprint+7, r16
	ldd	r16, Y+pcb_start+2
	sts	pprint+8, r16
	ldd	r16, Y+pcb_start+3
	sts	pprint+9, r16
	std	Z+ucb_diskaddr+0, zero
	std	Z+ucb_diskaddr+1, zero
	std	Z+ucb_imgptr+0, yl
	std	Z+ucb_imgptr+1, yh
	ldi	r16, (1<<pcb__att)
	std	Y+pcb_status, r16
	sts	pprint+10, r18
	call	print
	.db	"Attach Partition", CR, LF
	.db	"  Unit control block ......:0x", 0x81, 0x80, CR, LF
	.db	"  Partition control block .:0x", 0x83, 0x82, CR, LF
	.db	"  Volume type .............;0x", 0x84, CR, CR, LF
	.db	"  Volume status ...........:0x", 0x85, CR, CR, LF
	.db   	"  Start sector ............:0x", 0x89, 0x88, 0x87, 0x86, CR, LF
	.db	"  Decrement unit counter ..:0x", 0x8a, CR, LF, 0
	adiw	xh:xl, unitsize
	dec	r18
	breq	AnalyzePartSuccess
	rjmp	AnalyzePartNext

AnalyzePartSuccess:
	ldi	r24, IO_SUCCESS		; Return no error
	clc
AnalyzePartExit:
	pop	zh
	pop	zl
	pop	yh
	pop	yl
	pop	xh
	pop	xl
	pop	r23
	pop	r22
	pop	r20
	pop	r19
	pop	r18
	pop	r17
	pop	r16
	pop	r4
	ret
;--------------------------------------------------------------------------
;
;
;
AnalyzeExtended:
	ldd	r16, Y+P_Flag
	sbrc	r16, Part_Ext		; Did we have a primary extended partition?
	rjmp	AnalyzeExtende010	; Yes
;
;	This is the first extended partition. Note that the starting sector
;	number of the first primary extension needs to be added to all partitions
;	described in any extended partition, inlcuding the extended partitions.
;	Therefore we need to remember it for later use and at the same time
;	this is the offset to the first extended peration. The offset is stored
;	in the control block at P_Cluster for later use.
;	Extended partitions can be linked, i.e. it is possible to have another
;	extended partition in a partition table of an extend partition. So if we
;	have found an extended partition we save the starting sector number
;
	ori	r16, (1<<Part_Ext) | (1<<Part_Next)
	std	Y+P_Flag, r16		; We have seen the master extended partition
	ldd	r16, Z+M_PartStart+0
	std	Y+P_Cluster+0, r16	; Remember the extended parititon offset
	sts	pextended+0, r16	; Remember the global extended parititon offset
	ldd	r16, Z+M_PartStart+1
	std	Y+P_Cluster+1, r16
	sts	pextended+1, r16
	ldd	r16, Z+M_PartStart+2
	std	Y+P_Cluster+2, r16
	sts	pextended+2, r16
	ldd	r16, Z+M_PartStart+3
	std	Y+P_Cluster+3, r16
	sts	pextended+3, r16
	ret
;
AnalyzeExtende010:
	ori	r16, (1<<Part_Next)	; We have a next extended partition to analyze
	std	Y+P_Flag, r16

	lds	r16, pextended+0
	ldd	r17, Z+M_PartStart+0
	add	r16, r17
	std	Y+P_Cluster+0, r16
	lds	r16, pextended+1
	ldd	r17, Z+M_PartStart+1
	adc	r16, r17
	std	Y+P_Cluster+1, r16
	lds	r16, pextended+2
	ldd	r17, Z+M_PartStart+2
	adc	r16, r17
	std	Y+P_Cluster+2, r16
	lds	r16, pextended+3
	ldd	r17, Z+M_PartStart+3
	adc	r16, r17
	std	Y+P_Cluster+3, r16
	ret
;--------------------------------------------------------------------------
;
;	Input
;		part1start	Start Sector of FAT volume to mount
;		Y points to paramter block
;
;	This is not ready to be used, we need to use another
;	buffer to read the VBR (volume boot record) as we
;	must not modify the parameter block or the partition
;	table in the MBR buffer.
;
MountVolumeFAT:
	ldi	zl, low(pbuffer)	; Get address of data buffer to use
	ldi	zh, high(pbuffer)	; 
	std	Y+P_Address+0, zl	; Save it in the parameter block
	std	Y+P_Address+1, zh
	lds	xl, pcbqueue+0
	lds	xh, pcbqueue+1
	rjmp	MountVolume020
MountVolume010:	
	ldd	r16, Z+pcb_status
	sbrc	r16, pcb__fat
	rjmp	MountVolume030
MountVolume015:
	ldd	xl, Z+0
	ldd	xh, Z+1
MountVolume020:
	movw	zh:zl, xh:xl
	cp	zl, zero
	cpc	zh, zero
	brne	MountVolume010
	ldi	r24, 0xfb		; No FAT Volume
	rjmp	MountVolumeError
;
MountVolume030:
	ldd	r16, Z+pcb_id
	cpi	r16, 'C'
	brne	MountVolume015
	ldi	r16, 1<<MBRok		; 
	sts	VolStatus, r16		; Set VolStatus MBR read ok
	ldd	r16, Z+pcb_start+0
	std	Y+P_Sector+0, r16
	sts	part1start+0, r16
	ldd	r16, Z+pcb_start+1
	std	Y+P_Sector+1, r16
	sts	part1start+1, r16
	ldd	r16, Z+pcb_start+2
	std	Y+P_Sector+2, r16
	sts	part1start+2, r16
	ldd	r16, Z+pcb_start+3
	std	Y+P_Sector+3, r16
	sts	part1start+3, r16
	call	SD_CARD_READ		; Read VBR
	ldi	r24, 0			; Assume no error
	brcc	MountVolumeVBR
;
;	Error Exit
;
	ldi	r24, 0xfe		; Error reading VBR
MountVolumeError:
	sec
	ret				; Exit with CS
MountVolumeMagic:
	ldi	r24, 0xfc		; Invalid VBR
	rjmp	MountVolumeError
;
MountVolumeVBR:
	lds	r16, pbuffer+510
	cpi	r16, 0x55
	brne	MountVolumeMagic
	lds	r16, pbuffer+511
	cpi	r16, 0xaa
	brne	MountVolumeMagic
	lds	r16, VolStatus
	ori	r16, 1<<VBRok
	sts	VolStatus, r16		; VBR read ok
	ldd	zl, Y+P_Address+0
	ldd	zh, Y+P_Address+1
	ldd	r16, Z+V_BytesPSect+0	; Bytes per sector
	sts	bytespsect+0, r16
	ldd	r16, Z+V_BytesPSect+1
	sts	bytespsect+1, r16
	ldd	r16, Z+V_ReservedSect+0	; Number of reserved sectors
	sts	reservdsect, r16
	ldd	r16, Z+V_ReservedSect+1
	sts	reservdsect+1, r16
	ldd	r16, Z+V_PartStart+0	; Sectores before this partition
	sts	sectbefore, r16
	ldd	r16, Z+V_PartStart+1
	sts	sectbefore+1, r16
	ldd	r16, Z+V_PartStart+2
	sts	sectbefore+2, r16
	ldd	r16, Z+V_PartStart+3
	sts	sectbefore+3, r16
	ldd	r16, Z+V_PartSize+0	; Size of this partition
	sts	sectpart, r16
	ldd	r16, Z+V_PartSize+1
	sts	sectpart+1, r16
	ldd	r16, Z+V_PartSize+2
	sts	sectpart+2, r16
	ldd	r16, Z+V_PartSize+3
	sts	sectpart+3, r16
	ldd	r16, Z+V_NumFATs	; Number of FAT's
	sts	numfat, r16
	ldd	r16, Z+V_SecPClust	; Sectors per cluster
	sts	sectperclst, r16
	ldd	r16, Z+V_SecPFAT+0	; Sectors per FAT
	sts	sectperfat, r16
	ldd	r16, Z+V_SecPFAT+1
	sts	sectperfat+1, r16
	sts	sectperfat+2, zero
	sts	sectperfat+3, zero
	lds	xl, sectperfat+0
	lds	xh, sectperfat+1	; Check number of sectors per FAT
	sbiw	xh:xl, 0
	brne	MountVolume16		; if it's not zero then we have FAT16
	lds	r16, VolStatus		;
	ori	r16, 1<<FAT32
	sts	VolStatus, r16		; else we have a FAT32 volume
	ldd	r16, Z+V_SecPFAT32+0
	sts	sectperfat+0, r16
	ldd	r16, Z+V_SecPFAT32+1
	sts	sectperfat+1, r16
	ldd	r16, Z+V_SecPFAT32+2
	sts	sectperfat+2, r16
	ldd	r16, Z+V_SecPFAT32+3
	sts	sectperfat+3, r16
MountVolume16:
	lds	r16, part1start+0	; DataStart = PartitionStart + ReservedSectors
	lds	r17, reservdsect+0
	add	r16, r17
	sts	fat1start, r16
	lds	r16, part1start+1
	lds	r17, reservdsect+1
	adc	r16, r17
	sts	fat1start+1, r16
	lds	r16, part1start+2
	adc	r16, zero
	sts	fat1start+2, r16
	lds	r16, part1start+3
	adc	r16, zero
	sts	fat1start+3, r16

	lds	r16, fat1start+0
	sts	datastart+0, r16
	lds	r16, fat1start+1
	sts	datastart+1, r16
	lds	r16, fat1start+2
	sts	datastart+2, r16
	lds	r16, fat1start+3
	sts	datastart+3, r16
	lds	r18, numfat		; + NumFAT * SectorsPerFAT
MountVolumeData:
	lds	r16, datastart+0
	lds	r17, sectperfat+0
	add	r16, r17
	sts	datastart, r16
	lds	r16, datastart+1
	lds	r17, sectperfat+1
	adc	r16, r17
	sts	datastart+1, r16
	lds	r16, datastart+2
	lds	r17, sectperfat+2
	adc	r16, r17
	sts	datastart+2, r16
	lds	r16, datastart+3
	lds	r17, sectperfat+3
	adc	r16, r17
	sts	datastart+3, r16
	dec	r18
	brne	MountVolumeData	
	lds	r16, VolStatus
	sbrc	r16, FAT32
	rjmp	MountVolume32		; For FAT32 we are done

	lds	r16, datastart+0	; For FAT16 this is where the Root Directory Starts
	sts	rootdir, r16
	lds	r16, datastart+1
	sts	rootdir+1, r16
	lds	r16, datastart+2
	sts	rootdir+2, r16
	lds	r16, datastart+3
	sts	rootdir+3, r16
	ldd	xl, Z+V_EntriesRootD+0
	ldd	xh, Z+V_EntriesRootD+1	; Get number of entries in root DIR
	
	ldi	r18, 5		; Multiply by 32 = 2^5
MountVolumeRoot:
	add	xl, xl
	adc	xh, xh
	dec	r18
	brne	MountVolumeRoot
	
	lds	r16, bytespsect+0
	add	xl, r16
	lds	r16, bytespsect+1
	adc	xh, r16
	sbiw	xh:xl, 1		; Add bytes per sectors -1
	lsr	xh			; Devide by 512 gives number of
	sts	dirsectors, xh		; Sectors in Root Directory
	lds	r16, datastart+0
	add	r16, xh
	sts	datastart+0, r16
	lds	r16, datastart+1
	adc	r16, zero
	sts	datastart+1, r16
	lds	r16, datastart+2
	adc	r16, zero
	sts	datastart+2, r16
	lds	r16, datastart+3
	adc	r16, zero
	sts	datastart+3, r16	; Data Start of FA16
	clc
	ret

MountVolume32:
	sts	dirsectors, zero	; Invalidate for FAT32 and instead
	ldd	r16, Z+V_FAT32RootClus+0; save the first cluster of the root
	sts	rootdir+0, r16		; directory at as value for rootdir
	ldd	r16, Z+V_FAT32RootClus+1
	sts	rootdir+1, r16
	ldd	r16, Z+V_FAT32RootClus+2
	sts	rootdir+2, r16
	ldd	r16, Z+V_FAT32RootClus+3
	sts	rootdir+3, r16
	clc
	ret
;--------------------------------------------------------------------------
;
;	This routine is called to match the partition against a list of valid
;	drive types. In the case a valid drive type could be mapped to the
;	partition a partition control block is added to the partition queue.
;
;	Z	Partition Table Entry
;	Y	Parameterblock
;	r24	PCB Status
;
;--------------------------------------------------------------------------
;
;	Alternate entry for FAT-12 drives, note that this only works for 
;	small partitions which are 32Mbyte or less in size. This entry is
;	used to update the partition table of a newly formatted SD-Card and
;	changes the partition type from FAT-12 to Linux. If the magic
;	key 42 is given to the Q (mount volumes) command the partition
;	type is change else the partition will not be mounted and the
;	partition table will not be updated.
;
QueuePartImage:
	push	yl
	push	yh
;
;
;
	mov	r20, r24		; save PCB status
	ldd	r22, Z+M_PartSize+0
	ldd	r23, Z+M_PartSize+1
	ldd	r24, Z+M_PartSize+2
	ldd	r25, Z+M_PartSize+3
	rcall	FindDrvEntry		; r25:r24 = Drive Table Entry
	adiw	r25:r24, 0		;
	sbrs	r20, pcb__fat		;
	breq	QeueuPartExit
	movw	r23:r22, r25:r24	; save Drive Table Entry
;
;	Allocate a partition control block
;
	ldi	r24, low(pcb_len)
	ldi	r25, high(pcb_len)
	call	malloc
	brcs	QeueuPartExit
;
;
;
	ldd	r16, Y+P_Sector+0	; Get start of primary or extended partition
	ldd	r17, Y+P_Sector+1
	ldd	r18, Y+P_Sector+2
	ldd	r19, Y+P_Sector+3
;
;	We no longer need the address of the IO control block
;
	movw	Y, r25:r24		; Y = Partition Control Block
	std	Y+pcb_psector+0, r16
	std	Y+pcb_psector+1, r17	
	std	Y+pcb_psector+2, r18	
	std	Y+pcb_psector+3, r19	

	ldd	r24, Z+M_PartStart+0	; Add start sector
	add	r16, r24
	ldd	r24, Z+M_PartStart+1
	adc	r17, r24
	ldd	r24, Z+M_PartStart+2
	adc	r18, r24
	ldd	r24, Z+M_PartStart+3
	adc	r19, r24

	std	Y+pcb_start+0, r16	; Save start sector in volume information block
	std	Y+pcb_start+1, r17
	std	Y+pcb_start+2, r18
	std	Y+pcb_start+3, r19

	std	Y+pcb_drvtab+0, r22
	std	Y+pcb_drvtab+1, r23
	;
	;	Assume status in r20
	;
	std	Y+pcb_status, r20
	cpi	r20, (1<<pcb__fat)
	brne	QueuePart010
	;
	;	FAT Partition, gets a Drive Letter as identifier
	;
	lds	r16, pcbdisk
	std	Y+pcb_id, r16		; set partition id
	inc	r16
	sts	pcbdisk, r16
QueuePart010:
	cpi	r20, 0
	brne	QueuePart020
	;
	;	Image get the pcb counter as identifier
	;
	lds	r16, pcbcount
	inc	r16
	sts	pcbcount, r16
	std	Y+pcb_id, r16		; set partition id
QueuePart020:
	cpi	r20, (1<<pcb__fat12)
	brne	QueuePart030
	;
	;	Idle partition get the lower byte index to parititon table as status
	;	and the pcb counter as identifier
	;
	std	Y+pcb_status, r20
	std	Y+pcb_poff, zl
	lds	r16, pcbcount
	inc	r16
	sts	pcbcount, r16
	std	Y+pcb_id, r16		; set partition id
QueuePart030:
	rcall	InsertPcb
QeueuPartExit:
	pop	yh
	pop	yl
	ret
;--------------------------------------------------------------------------
;
;	Y	Partition Control Block
;
InsertPcb:
	push	xl
	push	xh
	push	zl
	push	zh
	ldi	zl, low(pcbqueue)
	ldi	zh, high(pcbqueue)

InsertPcb010:	
	ldd	xl, Z+0
	ldd	xh, Z+1
	sbiw	xh:xl, 0
	breq	InsertPcb020
	movw	zh:zl, xh:xl
	rjmp	InsertPcb010

InsertPcb020:
	std	Y+pcb_queue+0, zero
	std	Y+pcb_queue+1, zero
	std	Z+0, yl
	std	Z+1, yh
	pop	zh
	pop	zl
	pop	xh
	pop	xl
	ret

;--------------------------------------------------------------------------
;
;	Find drive entry in drvtab using the size of the partition
;	or file in blocks
;
;	Input
;
;	r22,r23,r24,r25	Size in blocks
;
;	Output
;
;	r25:r24		Pointer to to entry in drvtab, 0 if no entry found
;
;	Registers
;
;	r16
;
FindDrvEntry:
	push	zl
	push	zh

	ldi	zl, low(2*drvtab)
	ldi	zh, high(2*drvtab)

FindDrvEntryNext:
	lpm	r16, Z+			; Compare size of drive in drvtab
	cp	r22, r16		; with partition size
	lpm	r16, Z+	
	cpc	r23, r16
	lpm	r16, Z+	
	cpc	r24, r16
	lpm	r16, Z+	
	cpc	r25, r16

	breq	FindDrvEntryFound
	adiw	Z, drv_len - drv_sec	; next entry in drvtab

	lpm	r16, Z			; end of table?
	cpi	r16, 0xff
	brne	FindDrvEntryNext	; no try next entry
	clr	r24
	clr	r25
	pop	zh
	pop	zl
	ret

FindDrvEntryFound:
	movw	r25:r24, zh:zl		; get current pointer to drv_sec
	sbiw	r25:r24, drv_sec	; Adjust to start of entry
	pop	zh
	pop	zl
	ret

;--------------------------------------------------------------------------
;
;	The Drive Table has a list of parameters to identify the drive when
;	a file or a partition is attached to a unit. 
;
;	Drives are identified by the size (in blocks a 512bytes), in order to
;	translate sector/track/cylinder to logical block number each drive
;	entry has the required parameters. A type field is available which
;	can then be used be the IO routines (e.g. to identify the drive to
;	the operating system, e.g. RL01/02 or RM03/05 etc.) it also has a 
;	4 character drive name. 
;	
;	The table must end with a word that contains -1 (it is very unlikely
;	that the lower 16-bit of a disksize are equal to this number)
;
;--------------------------------------------------------------------------
recordstart	drv
record		drv, siz, 4
record		drv, sec, 1
record		drv, trk, 1
record		drv, cyl, 2
record		drv, typ, 2
record		drv, nam, 4
recordend	drv, len
drvtab:
;
;		RL01	40	2	256	5Mbyte
;
	.dw	lwrd(40*2*256/2)	; Sector Size is only 256bytes!
	.dw	hwrd(40*2*256/2)	;
	.db	40, 2			; Sector/Track, Track/Cylinder
	.dw	256			; Cylinders
	.db	0, 0			; Drive Type
	.db	"RL01"			; Name
;
;		RL02	40	2	512	10Mbyte
;
	.dw	lwrd(40*2*512/2)	; Sector Size is only 256bytes!
	.dw	hwrd(40*2*512/2)
	.db	40, 2
	.dw	512
	.db	(1<<DL_RL02), 0		; Drive Type
	.db	"RL02"
;
;		RL03	64	2	512	32Mbyte
;
	.dw	lwrd(64*2*512)
	.dw	hwrd(64*2*512)
	.db	64, 2
	.dw	512
	.db	0xf0, 0			; Drive Type
	.db	"RL03"
;
;		RD54	17	15	1225 159.9Mbyte
;
	.dw	lwrd(17*15*1225+1)	; 
	.dw	hwrd(17*15*1225+1)
	.db	0, 0			; LBA
	.dw	0
	.db	0xf0, 0			; Drive Type
	.db	"RD54"
;
;
;
	.dw	-1
