;--------------------------------------------------------------------------
;
;	Core RLV12 Emulator routines
;
;
;	2018-03-25	For logging move the drive select bits to CSRL bits 4 and 5, 
;			these bits are just a copy of bits 0 and 1 of BAE.
;	2018-12-08	Version 2.0 based on new Bridge Design. See CPLD.
;	2018-12-23	Include seek into general rlv12 file
;	2019-01-05	In case the drive is not ready do not just return drive
;			error but call a special error handling routine, especially
;			read header and get status need to be more sophisticated
;	2019-01-05	Also handle ucb_rdy in seek
;	2019-01-12	Include all RL01/02 specific
;	2019-07-26	New Logical2Physical (moved input paramter to P_Cluster)
;			We need to translate each block via Logical2Physical when
;			the disk image is a file as we never know if we advance to
;			the next fragment if we need to read more than one block
;			rlv12_rwsetup now makes sure that Z points to the ucb for
;			the rest of processing and added rlv12_rwnextsector (which
;			requires ucb_file flag on ucb_status therefore we now keep
;			the ucb in Z)
;	2019-11-21	Add new CPLD interface as option
;	2019-11-28	Centralized DMA routines
;	2020-12-25	New RLV12 Emulator for PDP-11/Hack
;--------------------------------------------------------------------------
	.cseg
;;;------------------------------------------------------------------------
;;;
;;;	Called by PC_INIT with interrupts disabled
;;;
rlv12_reset:
	ldi	r16, (1<<CSR_CRDY)	;;; Controller is ready
	sbi	b_CRDY
	sts	CSRL, r16
	sts	CSRH, zero
	sts	BARL, zero
	sts	BARH, zero
	sts	DARL, zero
	sts	DARH, zero
	sts	MPR_Fifo+0, zero
	sts	MPR_Fifo+1, zero
	sts	BAEL, zero
	sts	BAEH, zero
	sts	unittable+unitsize*0+ucb_diskaddr+0, zero	; Reset drives disk address
	sts	unittable+unitsize*0+ucb_diskaddr+1, zero	; 
	sts	unittable+unitsize*1+ucb_diskaddr+0, zero	; 
	sts	unittable+unitsize*1+ucb_diskaddr+1, zero	; 
	sts	unittable+unitsize*2+ucb_diskaddr+0, zero	; 
	sts	unittable+unitsize*2+ucb_diskaddr+1, zero	; 
	sts	unittable+unitsize*3+ucb_diskaddr+0, zero	; 
	sts	unittable+unitsize*3+ucb_diskaddr+1, zero	; 
	ret
;--------------------------------------------------------------------------
;
;
;
rlv12_init:
	sts	unittable+unitsize*0+ucb_status, zero	; Invalidate DL0: Partition
	sts	unittable+unitsize*1+ucb_status, zero	; Invalidate DL1: Partition
	sts	unittable+unitsize*2+ucb_status, zero	; Invalidate DL2: Partition
	sts	unittable+unitsize*3+ucb_status, zero	; Invalidate DL3: Partition
	ret
;--------------------------------------------------------------------------
;
;	Before we can call the rlv12 emulator we need to save "all"
;	registers.
;
rlv12_sub:
	push	r16
	push	r17
	push	r18
	push	r19
	push	r20
	push	r24
	push	r25
	push	xl
	push	xh
	push	yl
	push	yh
	push	zl
	push	zh
	sts	log_pbnflag, zero	; Assume no PBN to log
;+
;	First we setup all necessary information to call the appropriate 
;	function routine. At the same time we also initiate the log message
;	buffer which is later copied to the circular loggging buffer. Each
;	call to rlv12_sub creates a 8 bytes message. At the beginning the
;	message buffer is initialised with the following values. Later
;	each action routine may change the values as required for logging.
;
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	| 0 | 0 |A21|A20|A19|A18|A17|A16| 0 | IE|DS1|DS0|FC2|FC1|FC0|DRY|
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	|A15|A14|A13|A12|A11|A10| A9| A8| A7| A6| A5| A4| A3| A2| A1| A0|
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	|                              DAR                              |
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	|                              MPR                              |
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;
;	CSR
;	
;	+---+---+---+---+---+---+---+---++---+---+---+---+---+---+---+---+
;	|ERR|DE |E3 |E2 |E1 |E0 |DS1|DS0||CRY|IE |A17|A16|F2 |F1 |F0 |RDY|
;	+---+---+---+---+---+---+---+---++---+---+---+---+---+---+---+---+

	sts	rlv12_error, zero	; Assume no errors
	lds	yl, CSRH
	andi	yl, driveselect
	sts	CSRH, yl
	swap	yl			; convert to UCB index, each UCB is 16 bytes
	mov	zh, yl			; keep a copy
	clr	yh			; convert UCB index to UCB address
	subi	yl, low(-unittable)	; 
	sbci	yh, high(-unittable)	; 
	ldd	r16, Y+ucb_status	; Get Volume Status to T-Bit
	bst	r16, ucb__rdy		; 
	lds	r16, CSRL
	bld	r16, CSR_DRVRDY	; Update CSR with DRDY bit	
	sts	CSRL, r16
	mov	zl, r16			; keep a copy
;
;	Initialise the logging message buffer with the registers content
;	zh	has now DS1:0 in bits 4:5, all other bits are cleared
;	zl	has now the low byte of CSR including the drive ready
;
	cbr	zl, (1<<CSR_BA17) | (1<<CSR_BA16) ; remove A16:17 
	or	zl, zh			; insert DS
	bld	zl, 0			; Copy volume status to logging 
	sts	log_message+0, zl	; Function Code, Drive Select, DRDY, IE
	lds	r16, BAEL
	sts	log_message+1, r16
	lds	r16, BARL
	sts	log_message+2, r16
	lds	r16, BARH
	sts	log_message+3, r16
	lds	r16, DARL
	sts	log_message+4, r16
	lds	r16, DARH
	sts	log_message+5, r16
	lds	r16, MPRL
	sts	log_message+6, r16
	lds	r16, MPRH
	sts	log_message+7, r16
;
;	Use the drive ready in bit 0 of zl as part of the jump table
;	index
;
	andi	zl, 0x0f		; Isolate function code and DRDY
	clr	zh			; it to jump table index
	subi	zl, low(-rlv12_table)	; 
	sbci	zh, high(-rlv12_table)	;
	icall				; Call Function subroutine
;--------------------------------------------------------------------------
;
;	Check if logging is active for the unit
;
rlv12_logging:
	lds	r16, CSRH		;
	andi	r16, driveselect	; Get Unit number
	brne	rlv12_log010
	sbis	GPIOR0, log__rl0	; Log for unit 0
	rjmp	rlv12_nologging
	rjmp	rlv12_logunit

rlv12_log010:
	dec	r16
	brne	rlv12_log020
	sbis	GPIOR0, log__rl1	; Log for unit 1
	rjmp	rlv12_nologging
	rjmp	rlv12_logunit

rlv12_log020:
	dec	r16
	brne	rlv12_log030
	sbis	GPIOR0, log__rl2	; Log for unit 2
	rjmp	rlv12_nologging
	rjmp	rlv12_logunit

rlv12_log030:
	sbis	GPIOR0, log__rl3	; Log for unit 3
	rjmp	rlv12_nologging
;
;	Reserve space in circular logging buffer, we do this with interrupts
;	disabled to avoid corrupted buffer pointer
;
rlv12_logunit:
	cli				;;;
	lds	zl, log_pointer+0	;;; Logging
	lds	zh, log_pointer+1	;;; Logging
	movw	xh:xl, zh:zl		;;;
	lognxt	xl, xh			;;; next logging buffer address
	sei				;;;
	
	ldi	xl, low(log_message)	; Logging
	ldi	xh, high(log_message)	; Logging
;
;	Copy the 8 byte log message to the circular logging buffer
;
	ld	r16, X+
	st	Z+, r16
	ld	r16, X+
	st	Z+, r16
	ld	r16, X+
	st	Z+, r16
	ld	r16, X+
	st	Z+, r16
	ld	r16, X+
	st	Z+, r16
	ld	r16, X+
	st	Z+, r16
	ld	r16, X+
	st	Z+, r16
	ld	r16, X+
	st	Z+, r16
	sbis	GPIOR0, log__pbn	; pbn logging requested
	rjmp	rlv12_nologging		; no
	lds	r16, log_pbnflag		
	tst	r16			; pbn logging pending
	breq	rlv12_nologging		; no
;
;	Log physical block number
;
	cli				;;;
	lds	zl, log_pointer+0	;;; Logging
	lds	zh, log_pointer+1	;;; Logging
	movw	xh:xl, zh:zl		;;;
	lognxt	xl, xh			;;; next logging buffer address
	sei				;;;

	ldi	xl, low(log_pbnmsg)	; PBN Logging
	ldi	xh, high(log_pbnmsg)	; PBN Logging
;
;	Copy the 8 byte log message to the circular logging buffer
;
	ld	r16, X+
	st	Z+, r16
	ld	r16, X+
	st	Z+, r16
	ld	r16, X+
	st	Z+, r16
	ld	r16, X+
	st	Z+, r16
	ld	r16, X+
	st	Z+, r16
	ld	r16, X+
	st	Z+, r16
	ld	r16, X+
	st	Z+, r16
	ld	r16, X+
	st	Z+, r16
rlv12_nologging:
;
;	Set CRDY and update the CSR in the CPLD
;
	cli;----------------------------;;;
	lds	zl, CSRL
	lds	zh, CSRH
	sbr	zl, (1<<CSR_CRDY)	;;; Mark Controller Ready
	lds	r16, rlv12_error	;;; Copy over error bits
	or	zh, r16		;;;
	sbi	b_CRDY			;;; Enable CPLD / MCU interface
;
;	Set Interrupt request in case the IE bit of the CSR is set.
;
	sbrc	zl, CSR_IE		;;; Interrupts enabled?
	sbi	b_IRQST			;;; yes then set interrupt
	sts	CSRL, zl
	sts	CSRH, zh
	sei;----------------------------;;;

	pop	zh			; Restore the registers used by
	pop	zl			; rvl12_sub
	pop	yh
	pop	yl
	pop	xh
	pop	xl
	pop	r25
	pop	r24
	pop	r20
	pop	r19
	pop	r18
	pop	r17
	pop	r16
	ret
;--------------------------------------------------------------------------
;
;	Jump Table to the function Routines note function code uses bits 1..3,
;	bit0 is the drive ready bit. In other words we have to routines per
;	function call, the first when the drive is not ready and the second
;	when the drive is ready.
;
;	Function Code	Description			
;	000		No Op (RL11) or Maintenance Mode (RLV11/RLV12)
;	001		Write Check			
;	010		Get Status			
;	011		Seek				
;	100		Read Header			
;	101		Write Data			
;	110		Read Data			
;	111		Read Data Without Header Check
;
rlv12_table:
	rjmp	rlv12_nodrive
	rjmp	rlv12_maintenance

	rjmp	rlv12_nodrive
	rjmp	rlv12_writecheck


	rjmp	rlv12_getstatus_nd
	rjmp	rlv12_getstatus
	
	rjmp	rlv12_nodrive
	rjmp	rlv12_seek

	rjmp	rlv12_readheader_nd
	rjmp	rlv12_readheader

	rjmp	rlv12_nodrive
	rjmp	rlv12_writedata

	rjmp	rlv12_nodrive
	rjmp	rlv12_readdata

	rjmp	rlv12_nodrive
	rjmp	rlv12_readnocheck
;
;	If no volume is associated with the drive index then we return a drive
;	error.
;
rlv12_nodrive:
	ldi	r17, 0xc0		; Set error bits: Drive Error
	sts	rlv12_error, r17
	ret

;--------------------------------------------------------------------------
;
;	
;
rlv12_autoboot:
	push	r16
	push	r17
	push	r18
	push	r19
	push	xl
	push	xh
	push	yl
	push	yh
	push	zl
	push	zh
	sts	autoboot, zero		; ack autoboot
	ldi	yl, low(unittable)	; get address of first unit
	ldi	yh, high(unittable)	
	ldd	r16, Y+ucb_status
	sbrc	r16, ucb__part		; is a partition attached
	rjmp	rlv12_autobootpart	; then boot from partition
	sbrc	r16, ucb__file		; is a file attached
	rjmp	rlv12_autobootfile	; then boot from file
	rjmp	rlv12_noautoboot
;
;	boot from partition
;
rlv12_autobootpart:
;	call	print
;	.db	CR, LF
;	.db	"Boot partition!", CR, LF, 0
;	call	redraw_1
	ldd	zl, Y+ucb_imgptr+0	; get address of partition control block
	ldd	zh, Y+ucb_imgptr+1
	ldi	yl, low(pcluster)	; get IO Parameter block
	ldi	yh, high(pcluster)
	ldd	r16, Z+pcb_start+0	; copy partition start
	ldd	r17, Z+pcb_start+1
	ldd	r18, Z+pcb_start+2
	ldd	r19, Z+pcb_start+3
	std	Y+P_Sector+0, r16	; to IO Parameter block, this is the boot
	std	Y+P_Sector+1, r17	; sector
	std	Y+P_Sector+2, r18
	std	Y+P_Sector+3, r19
	rjmp	rlv12_autoboot010	; cont
;
;	boot from file
;
rlv12_autobootfile:
;	call	print
;	.db	CR, LF
;	.db	"Boot diskimage!", CR, LF, 0
;	call	redraw_1
	ldd	zl, Y+ucb_imgptr+0	; get address of file control block
	ldd	zh, Y+ucb_imgptr+1
	movw	yh:yl, zh:zl		; copy fcb
	std	Y+P_Cluster+0, zero	; start with first sector of file
	std	Y+P_Cluster+1, zero
	std	Y+P_Cluster+2, zero
	std	Y+P_Cluster+3, zero
	call	Logical2Physical	; translate using fragmentation list of FCB
;
;	IO Parameter Block is now setup with the sector to be read into memory
;
rlv12_autoboot010:
	ldi	r16, low(pbuffer)	; get default IO buffer
	ldi	r17, high(pbuffer)
	std	Y+P_Address+0, r16	; set IO buffer address
	std	Y+P_Address+1, r17
	ldi	r16, led_time		; blinken lights
	sts	led_oneshot, r16
	sbi	b_LED
	call	SD_CARD_READ		; read sector
	brcc	rlv12_autoboot015
	rjmp	rlv12_noautoboot
rlv12_autoboot015:
;
;	Create a log message at this point only yh:yl need to be preserved
;
	cli				;;;
	lds	zl, log_pointer+0	;;; Logging
	lds	zh, log_pointer+1	;;; Logging
	movw	xh:xl, zh:zl		;;;
	lognxt	xl, xh			;;; next logging buffer address
	sei				;;;
	ldi	r16, 1
	std	Z+0, r16
	std	Z+6, r16		; Read one sector
	ldi	r16, log_autoboot
	std	Z+1, r16
	ldd	r16, Y+P_Sector+0
	ldd	r17, Y+P_Sector+1
	ldd	r18, Y+P_Sector+2
	ldd	r19, Y+P_Sector+3
	std	Z+2, r16
	std	Z+3, r17
	std	Z+4, r18
	std	Z+5, r19
	std	Z+7, zero
;
;	we have now the boot record in our IO buffer, while the PDP-11
;	executes the BR . (0777) instruction at address 0 we copy 
;	words 1..256 of the boot sector to PDP-11 address 2
;	

	ldi	r16, 2			; Memory Address 0000000002
	dmaaddr	r16, zero, zero
;	rcall	rlv12_setupdmaaddress
	ldd	xl, Y+P_Address+0
	ldd	xh, Y+P_Address+1
	ld	zl, X+			; keep hold of the first word
	ld	zh, X+			; in the boot record
	ldi	r18, 1			; need to transfer 255 words
rlv12_autoboot020:
	ld	r16, X+			; now copy words 1..256. via DMA
	ld	r17, X+			; to the PDP-11 memory
	dmawrt	r16, r17
	inc	r18			; 
	brne	rlv12_autoboot020
;
;	Last but not least we need to copy the first word of the boot
;	sector to address zero of the PDP-11 memory. This will overwrite
;	the BR . (0777) instruction the PDP-11 is continuously executing
;	with the first instruction of the boot sector and therefore start
;	the boot process
;
	dmaaddr	zero, zero, zero
	dmawrt	zl, zh			; Write first word to address zero.
rlv12_noautoboot:
	pop	zh
	pop	zl
	pop	yh
	pop	yl
	pop	xh
	pop	xl
	pop	r19
	pop	r18
	pop	r17
	pop	r16
	ret
;--------------------------------------------------------------------------
;
;	Not (yet) implemented
;
rlv12_maintenance:
	ret
;--------------------------------------------------------------------------
;
;	Write Check, Compare data on the disk with the data in memory
;
rlv12_writecheck:
	ldi	r24, 1			; DMA Read
	rcall	rlv12_rwsetup
	ldi	r17, led_time
	sts	led_oneshot, r17
	sbi	b_LED
	call	SD_CARD_READ		; 
;
;	Now we need to consider the following special cases
;
;	If read starts at an odd sector we need to skip half
;	of the block from the SDCARD
;
	lds	r17, DARL
	sbrs	r17, 0
	rjmp	rlv12_writechkdmaloop
	inc	xh			; Second half of block from SDCARD
	ldi	r20, 0x80		; has only 128 words

rlv12_writechkdmaloop:
	dmaread	r16, r17
	brcs	rlv12_writechktmo
	ld	r18, X+
	ld	r19, X+
	cp	r16, r18
	cpc	r17, r19
	brne	rlv12_writechk_error
	adiw	r25:r24, 1		; increment complement of requested words
	breq	rvl12_writechk_success	; if eq we are really done
	dec	r20			; do we have still words in our buffer
	brne	rlv12_writechkdmaloop	; Yes
	rcall	rlv12_rwnextsector
	ldi	r17, led_time
	sts	led_oneshot, r17
	sbi	b_LED
	call	SD_CARD_READ
	clr	r20			; 256 words
	ldd	xl, Y+P_Address+0	; Get the buffer address used by the
	ldd	xh, Y+P_Address+1	; SD_CARD_READ routine and continue
	rjmp	rlv12_writechkdmaloop	;

rvl12_writechk_success:
	rjmp	rlv12_rw_success
;
;	DMA Read Time-Out Handler
;
rlv12_writechktmo:
	ldi	r17, 0xa0		; Set error bits: NXM
	sts	rlv12_error, r17
	ret
;
;	On compare error we print a message to the MCU console and set the 
;	corresponding error bits in CSR
;
rlv12_writechk_error:
	sts	pprint+0, r16		; Value from PDP-11 Memory
	sts	pprint+1, r17
	sts	pprint+2, r18		; Value from SD-Card
	sts	pprint+3, r19
	sts	pprint+4, xl		; Address in Buffer
	sts	pprint+5, xh
	call	print
	.db	"wcheck-error at buffer address: 0x", 0x85, 0x84
	.db	"  pdp-11: 0x", 0x81, 0x80
	.db	" disk: 0x", 0x83, 0x82, CR, LF, 0

	ldi	r17, 0x88		; Set error bits: WCE
	sts	rlv12_error, r17
	ret
;--------------------------------------------------------------------------
;
;	Get drive status
;
;	Often the DL: driver calls get status three times. First it does
;	a read status with RST cleared to get the current error bits. Then
;	it does a get status with RST set to clear the error bits and then
;	checks the drive again with RST cleared to make sure the error bits
;	are reset.
;
;	DAR during get status
;
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	| x | x | x | x | x | x | x | x | 0 | 0 | 0 | 0 |RST| 0 | 1 | 1 |
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;
;	MPR during get status
;
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	|WDE|CHE|WL |SKT|SPE|WGE|VC |DSE|DT |HS |CO |HO |BH |STC|STB|STA|
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;
rlv12_getstatus_nd:
	lds	r16, DARL		; Check for valid request code
	cpi	r16, 0x0B		; Get Status with reset errors
	breq	rlv12_getnd010
	cpi	r16, 0x03		; Get Status
	breq	rlv12_getnd010
	;
	; Here we should return an error, but which?
	;
	ret
	
rlv12_getnd010:
	ldi	r16, 0x0e		; Unload heads
	rjmp	rlv12_get020	

rlv12_getstatus:
	lds	r16, DARL		; Check for valid request code
	cpi	r16, 0x0B		; Get Status with reset errors
	breq	rlv12_get010
	cpi	r16, 0x03		; Get Status
	breq	rlv12_get010
	;
	; Here we should return an error, but which?
	;
	ret
rlv12_get010:
	ldi	r16, 0x1d		; Assume RL01 and drive ready
	ldd	r17, Y+ucb_diskaddr	; Get current Disk address
	bst	r17, DAR_RW_HS		; Get current head selected
	bld	r16, MPR_GETS_HS	; Set selected head in status
	ldd	r17, Y+ucb_type		; Get Volume Type
	bst	r17, DL_RL02		; Copy RL02 bit
	bld	r16, MPR_GETS_DT	; Set drive type in status
;	sbrs	r17, 0			; Check if Volume is valid
;	ldi	r16, 0x1f		; Volume is not ready

rlv12_get020:
	sts	log_message+6, r16	; 
	sts	log_message+7, zero	; Logging
;
;	Note that when the PDP-11 reads the MPR the reads always go to the
;	FIFO and never to the wordcount register, which is a write only
;	register.
;
	sts	MPR_Fifo+0, r16	; Return status
	sts	MPR_Fifo+1, zero
	sts	MPR_Fifo+2, zero
	sts	MPR_Fifo+3, zero
	sts	MPR_Fifo+4, zero
	sts	MPR_Fifo+5, zero
	sts	MPR_Fifo+6, zero
	sts	MPR_Fifo+7, zero
	ret	
;--------------------------------------------------------------------------
;
;	Seek RLV12 Emulator routines
;
;	DAR during seek
;
;	+---+---+---+---+---+---+---+---++---+---+---+---+---+---+---+---+
;	|DF8|DF7|DF6|DF5|DF4|DF3|DF2|DF1||DF0| 0 | 0 |HS | 0 |DIR| 0 | 1 |
;	+---+---+---+---+---+---+---+---++---+---+---+---+---+---+---+---+
;
;	Disk Address during read/write operations
;
;	+---+---+---+---+---+---+---+---++---+---+---+---+---+---+---+---+
;	|CA8|CA7|CA6|CA5|CA4|CA3|CA2|CA1||CA0|HS |SA5|SA4|SA3|SA2|SA1|SA0|
;	+---+---+---+---+---+---+---+---++---+---+---+---+---+---+---+---+
;
;
rlv12_seek:
	ldd	zl, Y+ucb_diskaddr+0
	ldd	zh, Y+ucb_diskaddr+1
	sts	log_message+2, zl
	sts	log_message+3, zh	; Log previous cylinder
;
;	DAR during seek
;
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	|DF8|DF7|DF6|DF5|DF4|DF3|DF2|DF1|DF0| 0 | 0 |HS | 0 |DIR| 0 | 1 |
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;
	lds	xl, DARL
	lds	xh, DARH
	sbrs	xl, DAR_SEEK_DIR	; direction of seek
	rjmp	rlv12_seekout		; towards the boarder
;
;	DIR=1 head moves to a higher cylinder address
;
	andi	xl, 0x80		; Only keep cylinder bits
	add	zl, xl
	adc	zh, xh
	rjmp	rlv12_seekdone
;
;	DIR=0 head moves to a lower cylinder address
;
rlv12_seekout:
	andi	xl, 0x80		; Only keep cylinder bits
	sub	zl, xl
	sbc	zh, xh
;
rlv12_seekdone:
	std	Y+ucb_diskaddr+0, zl
	std	Y+ucb_diskaddr+1, zh	; Save new cylinder
	sts	log_message+6, zl
	sts	log_message+7, zh

	ldd	r17, Y+ucb_status	; Reset Seek Request
	cbr	r17, (1<<DL_DRSEEK)
	std	Y+ucb_status, r17
	ret
;--------------------------------------------------------------------------
;
;	Read Header
;
;	Fake the read header result by calculating the header from a RL02
;	from just the current DAR. After every read header we increment the
;	sector number respecting overflow from 39 to 0.
;
rlv12_readheader_nd:
	ldi	r16, 0xc4		; Set error bits: Drive Error+OPI
	sts	rlv12_error, r16
	ldi	r16, 0xff
	sts	MPR_Fifo+0, r16
	sts	MPR_Fifo+1, r16
	
	sts	MPR_Fifo+2, zero
	sts	MPR_Fifo+3, zero
	sts	MPR_Fifo+4, zero
	sts	MPR_Fifo+5, zero
	sts	MPR_Fifo+6, zero
	sts	MPR_Fifo+7, zero
	ret

rlv12_readheader:
	clr	r16
	clr	r17			; Init CRC
	ldd	xl, Y+ucb_diskaddr+0
	ldd	xh, Y+ucb_diskaddr+1
	crc	xl, r16, r17
	crc	xh, r16, r17
	crc	zero, r16, r17
	crc	zero, r16, r17
	sts	MPR_Fifo+0, xl
	sts	MPR_Fifo+1, xh
	sts	MPR_Fifo+2, zero
	sts	MPR_Fifo+3, zero
	sts	MPR_Fifo+4, r16
	sts	MPR_Fifo+5, r17
	sts	MPR_Fifo+6, zero
	sts	MPR_Fifo+7, zero
	sts	log_message+6, xl
	sts	log_message+7, xh
;
;	After each Read Header we increment the sector to make the
;	program think the disk spins really
;
	ldd	xl, Y+ucb_diskaddr+0
	mov	xh, xl
	andi	xl, 0x3F		; Isolate Sector Number
	andi	xh, 0xC0		; Isolate Cyl0 and HS
	inc	xl
	cpi	xl, 40
	brlo	rlv12_readheader010
	clr	xl			; Sectornumber goes from 0..39 only
rlv12_readheader010:
	or	xl, xh			; Insert Cyl0 and HS
	std	Y+ucb_diskaddr+0, xl
	ret
;--------------------------------------------------------------------------
;
;	Translate DAR into a sector number of the partition
;
;	The 16-bit disk address in DAR 
;        15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;	|CY8|CY7|CY6|CY5|CY4|CY3|CY2|CY1|CY0| HD| S5| S4| S3| S2| S1| S0|
;	+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
;		bit0..5		Sector		0..39
;		bit6		Head		0..1
;		bit7..15	Cylinder	0..511
;
;	needs to be translate into a logical block number. For this we multiply
;	the compound value of (cylinder,header) by 20 as there are 40 sectors 
;	of 256bytes per track on a RL01/02 which corresponds to 20 sectors of 
;	512bytes on a SD-Card. Block numbers start with head = 0 for the first 
;	20 blocks and continue with head = 1 for the next 20 blocks of a cylinder.
;
;	20 in binary is 0b10100, so we need to calculate
;
;	ccccccccch000000	cylinder and head in DAR
; x            10100		binary for 20 in decimal
;
;	this is the same as calculating the following value
;
;	00ccccccccch0000	ccccccccch shifted by 4 bits
; +	0000ccccccccch00	ccccccccch shifted by 2 bits
;
;	For this we take DAR, mask out the sector value and shift the result to
;	the right by two bits. Then we duplicate the result and shift the
;	copy again by two bits to the right. Finally we add them together and
;	have the desired value.
;
;	Then we need to add the sector/2 to the above result. Remember sectors
;	on a RL02 are only 256bytes. This gives the offset in sectors into the
;	partition that holds the disk image of a RL02. Later we need to check
;	whether the IO starts at an even or odd sector.
;
rlv12_dar2lbn:
;
;	Set DMA Address do this first as we use the T-Flag as direction bit
;
	push	r16
	push	r17
	push	r18
	push	r19
	movw	r19:r18, r25:r24
	andi	r18, 0xc0		; need only the LSB of CYL and the HS bits
	lsr	r19
	ror	r18
	lsr	r19
	ror	r18			; Devide by four
	movw	r17:r16, r19:r18	; Make copy
	lsr	r19
	ror	r18
	lsr	r19
	ror	r18			; Divide by four again
	add	r16, r18
	adc	r17, r19
	mov	r18, r24
	andi	r18, 0x3f		; Isolate Sector Number
	lsr	r18			; Sectors on RL02 are only 256bytes
	add	r16, r18
	adc	r17, zero
	movw	r25:r24, r17:r16
	pop	r19
	pop	r18
	pop	r17
	pop	r16
	ret

;--------------------------------------------------------------------------
;
;	Setup a read or write operation
;
;	On entry
;		r24		must be set to DMA direction (0=Write, 1=Read)
;		Y		points to the current RL0x volume control block
;
;	On exit	
;		Y		points to the parameter buffer with sector
;				and buffer address set for SD_CARD_READ/WRITE
;		X		points to the SD-Card buffer
;		r25:r24		2's complement of wordcount
;		r20		set to zero, a SD-Card Block has 256. words
;		CPLD reg.	DMA direction and DMA start address are set
;	
rlv12_rwsetup:
;
;	Set DMA Address do this first as we use the T-Flag as direction bit
;
	lds	r16, BARL
	bst	r24, 0			; Get DMA Direction
	bld	r16, 0			; Copy to DMA direction bit in DMA address
	lds	r17, BARH
	lds	r18, BAEL
	dmaaddr	r16, r17, r18
	ldi	r24, 0xff
	sts	log_pbnflag, r24
	lds	r24, DARL
	lds	r25, DARH
	rcall	rlv12_dar2lbn		; Convert DAR to LBN
	movw	r17:r16, r25:r24
	clr	r18
	clr	r19			; 32-bit sector of partition
;
;	After rwsetup Z will point to the ucb and Y will point to either the fcb
;	(file control block) or ioc (IO control block) for block IO
;	
	movw	zh:zl, yh:yl		; Copy ucb pointer (needed in "next-sector")
	ldd	yl, Z+ucb_imgptr+0	; Get pointer to disk image control block
	ldd	yh, Z+ucb_imgptr+1
	ldd	r20, Z+ucb_status	; Get the status
	sbrs	r20, ucb__file		; Is unit attached to a file?
	rjmp	rlv12_rwsetupp		; no its a partition
;
;	A file is attached, Y points to the fcb. We put the LBN to the P_Cluster
;	offset of the fcb (file control block) which is then translated to a physical
;	block number by using the fragment list attached to the file control block
;
	std	Y+P_Cluster+0, r16	; Set start LBN for read or write
	std	Y+P_Cluster+1, r17
	std	Y+P_Cluster+2, r18
	std	Y+P_Cluster+3, r19
	call	Logical2Physical	; Convert to PBN (pyhsical block number)
	rjmp	rlv12_rwsetupd
	
rlv12_rwsetupp:
	ldd	r20, Y+pcb_start+0	; Add start of partition
	add	r16, r20
	ldd	r20, Y+pcb_start+1
	adc	r17, r20
	ldd	r20, Y+pcb_start+2
	adc	r18, r20
	ldd	r20, Y+pcb_start+3
	adc	r19, r20

	ldi	yl, low(pcluster)	; Setup parameter block for general IO
	ldi	yh, high(pcluster)	; 
;	
	std	Y+P_Sector+0, r16	; Set start sector for read or write
	std	Y+P_Sector+1, r17
	std	Y+P_Sector+2, r18
	std	Y+P_Sector+3, r19
;
rlv12_rwsetupd:

	ldd	r16, Y+P_Sector+0
	ldd	r17, Y+P_Sector+1
	ldd	r18, Y+P_Sector+2
	ldd	r19, Y+P_Sector+3
	ldi	r20, log_pbn

	ldi	xl, 1
	sts	log_pbnmsg+0, xl
	sts	log_pbnmsg+1, r20
	sts	log_pbnmsg+2, r16
	sts	log_pbnmsg+3, r17
	sts	log_pbnmsg+4, r18
	sts	log_pbnmsg+5, r19
	sts	log_pbnmsg+6, xl	; Read one sector
	sts	log_pbnmsg+7, zero

	ldi	xl, low(pbuffer)	; 
	ldi	xh, high(pbuffer)	; 
	std	Y+P_Address+0, xl	; Set buffer address for SD-Card block
	std	Y+P_Address+1, xh	; 
	clr	r20			; one SD-Card block has 256 words.
;
;	Get the number of words requested. If more words are requested than would be
;	available on the current track from the start sector then we truncate to the
;	maximum wordcount possible. If this was the case we need to return HNF so the
;	device driver can continue the read/write request after a seek to the next
;	track has been made. This is an undocumented feature and has been verifed with
;	simh and at least the boot loader of RT-11 makes use of this behaviour. Note that
;	we calculate MAXWC and then overwrite MPR in case it is higher than the user
;	provided value of the wordcount in MPR. This value will later be used to calculate
;	the and bus address after the transferred bytes. This is not a problem as the
;	wordcount cannot be retrieved, reading MPR always returns values from the FIFO
;
	lds	r24, MPRL
	lds	r25, MPRH		; Get wordcount (2's complement) = wc
	lds	r17, DARL
	andi	r17, 0x3f
	subi	r17, 40			; Number of sectors left in current track
	clr	r16			; as a 2's complement number. Then we
	sec				; convert this into a maximum wordcount
	ror	r17			; that can be transferred before a SEEK
	ror	r16			; is required to switch tracks
	cp	r24, r16		; compare "wordcount" with "max wordcount"
	cpc	r25, r17		;
	brsh	rlv12_rwsetup010	; if less words request than possible -> ok
	mov	r24, r16
	mov	r25, r17		; only maxwc can be transferred
	sts	MPRL, r24
	sts	MPRH, r25		; used later to calculate the ending DMA address

	ldi	r17, 0x94		; Set error bits: Header Not Found
	sts	rlv12_error, r17
rlv12_rwsetup010:
	ret

;--------------------------------------------------------------------------
;
;	The RLV12 write command can write to any 256byte sector and it also can
;	write less then 256bytes. When writing less then 256bytes the RLV12 will
;	fill the rest with the value zero. The SD-Cards have a block size of
;	512bytes. Therefore it can happen that we need to first read the block
;	as only part of it is updated by the write request. MPR is loaded with the
;	2's complement of the wordcount to write, therefore we need to compare it
;	with the 2's complement of 128.
;	In case MPR is higher or same (between 177600 and 177777) we
;	only write one RL01/02 sector, which is half of a SD-Card block. This check
;	needs to be done always before we write a new SD-Card block. Note that 
;	a single write can write up to 5120 words as long as the RL01/02 sectors 
;	written belong to the same track. 
;
rlv12_writedata:
	ldi	r24, 1			; DMA Read
	rcall	rlv12_rwsetup		;
	lds	r17, DARL		; Does the write start at a SD-Card
	sbrc	r17, 0			; Block boundary
	rjmp	rlv12_write010		; -> no

rlv12_write000:
	ldi	r16, low(-128)		; Need to compare MPR with words per
	ldi	r17, high(-128)		; sector. Do it the clean way :-)
	cp	r24, r16		; Do we write more than 128 words?
	cpc	r25, r17		; That is MPR is lower than -128 (unsigned)
	brlo	rlv12_write030		; Just write the block
	rjmp	rlv12_write020		; Only first half of SD-Card block is written 
;
;	If the first sector to be written is an odd sector number then it starts
;	in the second half of a SD-Card block. Therefore we need to adjust the
;	start address and the number of words left to be written in a SD-Card
;	block.
;
rlv12_write010:
	ldi	r20, 0x80		; Write starts at the second half of
	inc	xh			; the SD-Card block.
;
;	Either write starts at second half of SD-Card Block or we only write
;	128 words or less to the first half therefore we first need to read 
;	the SD-Card block to the buffer before we transfer the data
;
rlv12_write020:
	ldi	r17, led_time
	sts	led_oneshot, r17
	sbi	b_LED
	call	SD_CARD_READ		; 
;
;	Now we transfer the data from PDP-11 memory to the Block buffer.
;
rlv12_write030:
	;
	;	New DMA
	;
	dmaread	r16, r17
	brcs	rlv12_writetmo
	st	X+, r16
	st	X+, r17
	adiw	r25:r24, 1		; One word less
	breq	rlv12_write040		; Done with the request
	inc	r20			; Another word in our SD-Card block buffer
	brne	rlv12_write030		; more left
;
;	We have reached the end of the SD-Card buffer, therefore we need to write
;	the data to the SD-Card and prepare for the next SD-Card block
;
	ldi	r17, led_time
	sts	led_oneshot, r17
	sbi	b_LED
	call	SD_CARD_WRITE
;
;	Update parameter block, i.e. increment current sector by 1.
;
	rcall	rlv12_rwnextsector
	clr	r20			; Once we are here we always start at the
	ldd	xl, Y+P_Address		; Get buffer address for SD-Card block
	ldd	xh, Y+P_Address+1	; 
	rjmp	rlv12_write000		; Do check for full or partial write of SD-Card block
;
;	The request has finished. If not a full sector has been written the rest
;	is filled with zero bytes. First we need to make sure we only fill the 
;	part that corresonds to the RL01/02 sector. Therefore we adjust r20
;	so it only accounts for one 128 word sector, regardless whether we have
;	started with the even or odd sector within the same SD-Card block
;
rlv12_write040:
	ori	r20, 0x80		; zero fill only to the next 128 word boundary
rlv12_write050:
;
;	Check we already reached the end of the SD-Card before writing zero as we
;	arrive here when wordcount has reached zero and before r20 would have been
;	incremented
;
	inc	r20			; need to increment r20 not done previously
	breq	rlv12_write060		; no more or no words to write
	st	X+, zero
	st	X+, zero		; Write one word with 0
	rjmp	rlv12_write050		; do for the rest of the buffer
;
;	Sector has now potentially be filled with zero value and the last SD-Card
;	block of this write request is ready to be written to the SD-Card
;
rlv12_write060:
	ldi	r17, led_time
	sts	led_oneshot, r17
	sbi	b_LED
	call	SD_CARD_WRITE		; Write block to SD-Card
	rjmp	rlv12_rw_success	; Done

rlv12_writetmo:
;
;	DMA was not successfull we assume bus time-out 
;
	sts	pprint+0, r17
	sts	pprint+1, xl
	sts	pprint+2, xh
	sts	pprint+3, r24
	sts	pprint+4, r25
	sts	pprint+5, r20
	call	print
	.db	0x0d,0x0a, "DMA Write Timeout ", 0x0d, 0xa
	.db	"    dmatmo......0x", 0x80, " ", 0x0d, 0x0a
	.db	"    dataptr.....0x", 0x82, 0x81, 0x0d, 0x0a
	.db	"    wourdcount..0x", 0x82, 0x81, 0x0d, 0x0a
	.db	"    blockcount..0x", 0x85, 0x0d, 0x0a, 0x00
	ldi	r17, 0xa0		; Set error bits: NXM
	sts	rlv12_error, r17
	ret
;--------------------------------------------------------------------------
;
;	Read Data
;
rlv12_readdata:
	ldi	r24, 0			; DMA Write
	rcall	rlv12_rwsetup
	sbi	b_LED
	ldi	r17, led_time
	sts	led_oneshot, r17
	call	SD_CARD_READ		; 
;
;	We need to consider the following special cases
;
;	-	If the software requested an odd start sector 
;	-	Wordcount (MPR) is not a multiple of 256
;
;	If read starts at an odd sector we need to skip half
;	of the block from the SDCARD
;
	lds	r17, DARL
	sbrs	r17, 0
	rjmp	rlv12_readdmaloop
	inc	xh			; Second half of block from SDCARD
	ldi	r20, 0x80		; has only 128 words

rlv12_readdmaloop:
	ld	r16, X+			; Fetch one word from buffer
	ld	r17, X+
	dmawrt	r16, r17		; Write the Word via DMA to PDP-11 Memory
	brcs	rlv12_readtmo
	adiw	r25:r24, 1		; increment complement of requested words
	breq	rlv12_rw_success	; if eq we are really done
	dec	r20			; do we have still words in our buffer
	brne	rlv12_readdmaloop	; Yes
	rcall	rlv12_rwnextsector
	sbi	b_LED
	ldi	r17, led_time
	sts	led_oneshot, r17
	call	SD_CARD_READ
	clr	r20			; 256 words
	ldd	xl, Y+P_Address		; Get buffer address of SD-Card block
	ldd	xh, Y+P_Address+1	; 
	rjmp	rlv12_readdmaloop	; and go
;
;	DMA Write Time-Out Handler
;
rlv12_readtmo:
	ldi	r17, 0xa0		; Set error bits: NXM
	sts	rlv12_error, r17
	ret

;--------------------------------------------------------------------------
;-
;	read or write data finished, we now need to update the DAR and BAR
;
rlv12_rw_success:
	lds	r24, MPRL
	lds	r25, MPRH
	lsl	r24			; Translate wordcount to bytecount
	rol	r25			; r25 holds also the number of RL01/02 sectors
	lds	r17, BARL		; update bus address
	sub	r17, r24		; Note wordcount was 2's complement
	sts	BARL, r17
	lds	r17, BARH
	sbc	r17, r25
	sts	BARH, r17
	lds	r17, BAEL
	sbci	r17, -1
	cli
	sts	BAEL, r17
	lds	zl, CSRL
	bst	r17, BAE_BA16		; update csrl with BA16, BA17
	bld	zl, CSR_BA16
	bst	r17, BAE_BA17
	bld	zl, CSR_BA17
	sts	CSRL, zl
	sei
;
;	in the setup routine we made sure that no more words will be transferred
;	than are left on a cylinder, if we read past the last sector, the sector
;	number obviously is now 40. which is the number the RLV12 effectively
;	returns, in this case the setup routine has set the error to HNF (header
;	not found)
;
	lds	r17, DARL
	sub	r17, r25		; r25 has number of sectors transferred 2's com
	sts	DARL, r17
	lds	zl, CSRH
	andi	zl, driveselect
	swap	zl
	clr	zh
	subi	zl, low(-unittable+unitsize*0)
	sbci	zh, high(-unittable+unitsize*0)
	std	Z+ucb_diskaddr+0, r17
	lds	r17, DARH
	std	Z+ucb_diskaddr+1, r17
	ret

;--------------------------------------------------------------------------
;
;	When reading or writing another block we need to either increment the
;	sector in case the unit is attached to a partition or we need to increment
;	the logical block number and then translate it to a physical block number.	
;
rlv12_rwnextsector:
	lds	r17, log_pbnmsg+6
	inc	r17
	sts	log_pbnmsg+6, r17	; Another block
	ldd	r17, Z+ucb_status
	sbrc	r17, ucb__file		; Is the unit attached to a file?
	rjmp	rlv12_rwnextsector010	; Yes so increment LBN and translate
	ldd	r17, Y+P_Sector+0	; Else we just need the next sector on the volume
;	add	r17, one
	subi	r17, byte1(-1)
	std	Y+P_Sector+0, r17
	ldd	r17, Y+P_Sector+1
;	adc	r17, zero
	sbci	r17, byte2(-1)
	std	Y+P_Sector+1, r17
	ldd	r17, Y+P_Sector+2
;	adc	r17, zero
	sbci	r17, byte3(-1)
	std	Y+P_Sector+2, r17
	ldd	r17, Y+P_Sector+3
;	adc	r17, zero
	sbci	r17, byte4(-1)
	std	Y+P_Sector+3, r17
	ret
	
rlv12_rwnextsector010:
	ldd	r17, Y+P_Cluster+0	; Else we need the next sector in the file
;	add	r17, one		; 
	subi	r17, byte1(-1)
	std	Y+P_Cluster+0, r17
	ldd	r17, Y+P_Cluster+1
;	adc	r17, zero
	sbci	r17, byte2(-1)
	std	Y+P_Cluster+1, r17
	ldd	r17, Y+P_Cluster+2
;	adc	r17, zero
	sbci	r17, byte3(-1)
	std	Y+P_Cluster+2, r17
	ldd	r17, Y+P_Cluster+3
;	adc	r17, zero
	sbci	r17, byte4(-1)
	std	Y+P_Cluster+3, r17
	call	Logical2Physical	; And translate it to a PBN using the fragment list
	ret				; attached to the FCB
	
;--------------------------------------------------------------------------
;
;	
;
rlv12_readnocheck:
	rjmp	rlv12_readdata		; In fact the same for a SD-Card

