.include	"m1284Pdef.inc"
.listmac
;--------------------------------------------------------------------------
;
;	RLV12 Emulator 2.0 for Q-Bus
;
;	Peter Schranz
;	peter (at) 5volts (dot) ch
;	Version 2.1 - 2022-04-12
;
;	This is a general clean-up version with substantial changes
;	not only to the code but also to the CPLD. There is no need
;	to further maintain any other releases of the RLV12 Emulator
;	-	All RLV12 Emulator for the PDP-11/Hack are obsolete
;		and are going to be replaced by the Disk Emulator
;	-	RVL12 Emulator 1.0 for Q-Bus will be abandonded
;	-	There will be only one RLV12 Emulator 2.0 for the
;		Q-Bus using the CPLD Interface 2.2
;
;	2022-04-19 start the clean-up for the following points
;	-	Register Cleanup, avoid predefined registers
;		1. char will be replaced to r24
;		2. eliminate count
;		3. eliminate temp
;	-	Routines are free to use r16-r25, xl, xh, zl and zh
;	-	Routines must preserve r0-r7, r12, r13 and must not use
;		r8-r11, r14, r15
;	-	Parameters are passed via r25:r24 or yh:yl, latter is not
;		conform to ABI but is often used in RLV12 Emulator
;	-	avoid using crcl and crch
;	-	as long as temp and count are defined avoid using r19 and r20
;	
;	2022-04-21 temp has been eliminated 
;	2022-04-22 char and count hava been eliminated
;	2022-04-22 activate readcmd with readcmdline instead entername
;	2022-05-06 Logging: Clean-Up of Code, new Message Format, new buffer
;		definitions and implement lollipop and circular shaped
;		buffers with new solution that only requires 2 cycles
;		regardless of logging buffer type
;	2022-05-15 crcl and crch eliminated, using new crc macro which
;		has the crc registers as parameter
;
;		IOStatus is now 4 bytes so we can return more info like
;		calculated and read CRC from SD_CARD_READ, crcl and crch
;		are now obsolete
;
;		shuffling routines, action routes are now in Qbridge11-
;		cli-action.asm, cli command routines in Qbridge11-cli.asm
;		and all logging are now in Qbridge11-logging.asm
;
;		MountVolume now saves all registers as required by the
;		serin->scheduler->sdcardinsert, so no registers except
;		r24 and r25 are modified by sdcardinsert and thus serin
;		readinit is now in Qbridge11-readinit.asm and saves all
;		registers that are not saved by MountVolume or could
;		conflict with parser scancommand
;
;		FindDrvEntry now uses ABI register convetions and gets
;		the size via r25:r24:r23:r22 and returns the pointer
;		in r25:r24
;	2024-05-15 Added new READ command, perform various clean-up
;
;--------------------------------------------------------------------------
.include "../include/macro-library-v1-1.asm"
.include "../include/FAT/fat-defs.asm"
.include "error.inc"

.equ	F_CPU		= 22118400	; Crystal clock

.equ	BAUD0		= 115200	; Baudrate USART0

.equ	RLVVECTOR	= 0160		; Default vector 0160 (octal)

;--------------------------------------------------------------------------
;
; Baud Rate Calculator
;
.equ	UBRR0_VAL   = ((F_CPU+BAUD0*8)/(BAUD0*16)-1)	; nearest value
.equ	BAUD0_REAL  = (F_CPU/(16*(UBRR0_VAL+1)))	; effective baudrate
.equ	BAUD0_ERROR = ((BAUD0_REAL*1000)/BAUD0-1000)	; error in 0/00
.if ((BAUD0_ERROR>10) || (BAUD0_ERROR<-10))
	.error " USART0: Effective baudrate is more then 1 percent off requested baudrate"
.endif
;--------------------------------------------------------------------------
;
;	RLV12 Assembler Option
;
#define cpld				; activates CPLD interface
#define loaduser			; user defined monitor load function
#define	rlv12	  22			; Set CPLD Interface Version
;
;	Conditionals and constants
;
.equ	debug		= 0		; interrupt counters
.equ	signal		= 0		; signalling with b_SIG
.equ	spi2xb		= 1		; the value of SPI2X bit

.equ	sdspi_i		= 101
.equ	sdspi_o		= 200
.equ	dmatmo		= 60		; DMA Timeout
.equ	led_time	= 10

;--------------------------------------------------------------------------
;
;	Register definitions and register convetion
;
;		  r0, r1		; may be destroyed by any function
;		  r2 to r10		; these registeres must be preserved
.def	zero 	= r11
;		  r12, r13, r14, r15	; these registeres must be preserved
;
;		  r16			; General register never saved 
;		  r17			; General register never saved 
;		  r18			; General register never saved 
;		  r19			; General register never saved 
;		  r20			; General register never saved 
;		  r21			; General register never saved 
;
;		  r22, r23		;;; Parameter or Temporary Register
;		  r24, r25		;;; Parameter or Temporary Register
;		  xl, xh		;;; Temporary Pointer
;		  yl, yh		;;; Parameter or Temporary Register
;		  zl, zh		;;; Temporary Pointer
;
;	Register Convention not yet strict but you can assume that for
;	the moment you may always destry r0, r1, r16 and r17, further
;	cleanup is required to allow r18, r19 and r20 to be also always
;	destroyed.
;
;	yh:yl	mostly used as pointer to pass a data structure as parameter
;	r25:r24	used as parameter and result for  functions that conform
;	to the ABI, like malloc, free, serin, serout, FindDriveEntry
;
;--------------------------------------------------------------------------
	.dseg
;
;	Memory usage
;
;	0x0100..0x0FFF	Buffers and Variables
;	0x1000..0x1FFF	Monitor and Test
;	0x2000..0x23FF	"ROM"
;	0x2400..0x2FFF	Heap
;	0x3000..0x3FFF	Logging Buffer
;	0x4000..0x40FF	Stack
;
;	Sector Buffers
;
pbuffer:	.byte	512		; General purpose sector buffer
pfatbuffer:	.byte	512		; Sector buffer for FAT sector
pdirbuffer:	.byte	512		; Sector buffer for directory sector
;
;	Text Buffers
;
plfilename:	.byte	256		; Long File Name Buffer
NameBuffer:	.byte	256		; Name Buffer for "CopyName"
Path:		.byte	256		; Current Working Path
InputBuffer:	.byte	256		; Input Buffer (Command Line Buffer)
HistoryBuffer:	.byte	256
;
;	Interrupt driven ring buffers for serial communication
;
txring:		.byte	256
rxring:		.byte	256
;
;	Variables:
;
INTQ_count:	.byte	1		; Number of DATI/O occured since last show
INTI_count:	.byte	1		; Number of IACK occured since last show
INIT_count:	.byte	1		; Number of BINIT occured since last show
INTB_count:	.byte	1		; Number of Busy interrupt occured 
autoboot:	.byte	1
log_pointer:	.byte	2		; current pointer into circular logging buffer
sd_cmd:		.byte	6		; 6 bytes for SD Card commands
sd_rsp:		.byte	6		; 6 bytes for SD Card responses
sd_status:	.byte	1
.equ	sd__reset_bp	= 0		; New nomenclature of bits
.equ	sd__ver2_bp	= 1
.equ	sd__init_bp	= 2
.equ	sd__ccs_bp	= 6
.equ	sd__reset_bm	= 0x01		; Card could be reset successfully
.equ	sd__ver2_bm	= 0x02		; If card is a version 2 card
.equ	sd__init_bm	= 0x04		; Card could be initialised
.equ	sd__ccs_bm	= 0x40		; Card Capacity Support
s_CDINS:	.byte	1		; debounce CD inserting card
s_CDINSD:	.byte	1		; insert CD detected
s_CDINSC:	.byte	1
s_CDREM:	.byte	1		; debounce CD removing card
s_CDREMD:	.byte	1		; remove CD detected
s_CDREMC:	.byte	1
;
;
;
txinptr:	.byte	1
txoutptr:	.byte	1
txcnt:		.byte	1
rxinptr:	.byte	1
rxoutptr:	.byte	1
rxcnt:		.byte	1
resetptr:	.byte	1
;
rlv12_error:	.byte	1
led_oneshot:	.byte	1
romstatus:	.byte	1
romaddress:	.byte	1		; Temporary storage for read ROM interrupt
debugreg:	.byte	32
debugcheck:	.byte	32
;--------------------------------------------------------------------------
;
;	Queue Header for malloc() and free()
;
heap:		.byte	4
.equ	heapstart	= 0x2400
.equ	heapsize	= 0x0C00

;--------------------------------------------------------------------------
;
;	Softrom Buffer
;
.equ	softrombuffer	= 0x2000
.equ	softromsize	= 0x0200
;--------------------------------------------------------------------------
;
;	Partition control block queue
;
pcbqueue:	.byte	2
pcbcount:	.byte	1
pcbdisk:	.byte	1
recordstart	pcb
record		pcb, queue, 2
record		pcb, status, 1
	;
	;	bits 0..3 are status flags
	;
	.equ	pcb__fat	=  0	; FAT 16 or FAT 32
	.equ	pcb__att	=  1	; Partition is attached
	.equ	pcb__fat32	=  2	; FAT Partition is FAT-32
	.equ	pcb__fat12	=  3	; FAT 12
record		pcb, id, 1		; Partition Number/ID
record		pcb, start, 4		; Start Sector
record		pcb, drvtab, 2		; Ppointer to Drive Tab Entry
record		pcb, psector, 4		; Sector that holds the paritition table
record		pcb, ptype, 1		; Partition Type
record		pcb, poff, 1		; Partition Table Entry Offset
recordend	pcb, len

;--------------------------------------------------------------------------
;
;	Logging 
;
.include "logging-defs.inc"

;--------------------------------------------------------------------------
;
;	Special characters
;

.equ	NULL	= 0
.equ    CR      = 015
.equ    LF      = 012
.equ    BELL    = 07
.equ    BKSP    = 010
.equ    TAB     = 011
.equ    EOF     = 032
.equ    ESC     = 033
.equ	SPACE	= 040
.equ    DEL     = 0177
.equ	CTRL_R	= 0x12			; redraw character
.equ	DELIM	= 0x2f			; Directory Delimiter 0x2f = / and 0x5c = \
;
;	Include controller specific definitions
;
.include "Qbridge11-rlv12-defs-v2.asm"
;--------------------------------------------------------------------------
;	IO Control Blocks
;--------------------------------------------------------------------------
;
;	The volume access primitives use fixed parameter blocks which are pointed
;	to by the 16-bit Y register (yl:yh). The parameter blocks have a fixed 
;	header that is valid for all data structures accessed. Depending on the
;	use of the parameter block fields are appended.
;
;--------------------------------------------------------------------------
recordstart	P
record		P, Cluster, 4
record		P, Sector, 4
record		P, Address, 2
record		P, IOStatus, 4
record		P, Duration, 2
record		P, Flag, 1
.equ	D_Linked	= 0		; Directory is a linked list of clusters
.equ	D_Long		= 1		; We have a long filename
.equ	Part_Next	= 5		; Analyze next extended partition table
.equ	Part_Fat	= 6		; FAT Volume Mapped
.equ	Part_Ext	= 7		; Current MBR has Extended Partition

record		P, NumSect, 1
recordend	P, Size

pdirstartc:	.byte	4		; Used in fdisk
pextended:	.byte	4		; Sector of primary extended partition

;--------------------------------------------------------------------------
;
;	Parameterblock used to read FAT, we always keep the last FAT
;	sector in a dedicated buffer to improve the performance when
;	building the fragmentlists note the fat control block has no
;	extension and is used as the prototype for the control block
;	header for IO control blocks, it is important that this never
;	chagnes
;
pfatcluster:	.byte	P_Size
;
;--------------------------------------------------------------------------
;
;	Paramterblock for general IO to read/write individual sectors like
;	MBR, VBR, Extended Partition, and general Sectors
;
pcluster:	.byte	P_Size		; current cluster / next sector
;
;--------------------------------------------------------------------------
;
;	Directory funtions, this parameter block is used to access
;	directories. Normally we only need one parameter block for this
;	purpose as we need it only to find files by name.
;
recordcont	P, Size
record		P, Count, 1
record		P, Res1, 1
record		P, DirPointer, 2
record		P, LongFileN, 2
record		P, StartClust, 4
record		P, SaveClust, 4
record		P, DirName, 2
recordend	P, Size2
;
pdircluster:	.byte	P_Size2		; active directory cluster
;
;--------------------------------------------------------------------------
;
;	Parameter Block for FILE IO	(File Control Block)
;
;--------------------------------------------------------------------------
;
recordstart	fcb
record		fcb, cluster, 4
record		fcb, sector, 4
record		fcb, address, 2
record		fcb, IOStatus, 4
record		fcb, Duration, 2
record		fcb, flag, 1
.equ	F_Readonly	= 0		; File is open read-only
.equ	F_Direct	= 1		; File is open for direct block IO
.equ	F_Sequential	= 2		; File is open for sequential access
.equ	F_Image		= 3		; File is a valid diskimage
record		fcb, secinclst, 1
record		fcb, drvtab, 2
.if ((fcb_drvtab) != (P_Size))
	.error " Invalid File IO Parameter Block, size mismatch!"
.endif
record		fcb, filename, 2
record		fcb, fraglist, 2
record		fcb, position, 4
record		fcb, filesize, 4
record		fcb, byteinsec, 2
recordend	fcb, length
;
recordstart	Fr
record		Fr, List, 2		; Linked List Head
record		Fr, Size, 4		; Size of Fragment (or 0 if end if list)
record		Fr, Start, 4		; Start sector of fragment on device
recordend	Fr, Entry		; Length of fragment list entry
;
;--------------------------------------------------------------------------
;	SD-Card control block
;--------------------------------------------------------------------------
;
;	To properly read information from the SD-Card we need very few data
;	which is extracted from the MBR and VBR and does not change once the
;	SD-Card is mounted. When mounting the SD-Card we read all partition
;	entries. Partitions that match the size of a PDP-11 volume are marked
;	as a volume and the first FAT (FAT-16 or FAT-32) partition is/will be
;	used for disk-images.
;
;--------------------------------------------------------------------------
;
;	To read the correct sector the following translations must be performed
;
;	RootDirSectors = RootDirEntries / 32;
;
;	FirstDataSector = BPB_ResvdSecCnt + (BPB_NumFATs * FATSz) + RootDirSectors;
;
;	FirstSectorofCluster = ((N^2) * BPB_SecPerClus) + FirstDataSector;
;
part1start:	.byte	4	; Startsector of 1st FAT partition: VBR
;
;	general info
;
bytespsect:	.byte	2	; Bytes per sector, we will only support 512bytes per sector
reservdsect:	.byte	2	; Reserved sectors normally FAT16: 1, FAT32: 32
;
;	to access the data areas
;
sectbefore:	.byte	4	; Number of sectors before this partition (same as part1start)
sectpart:	.byte	4	; Sectors in this partition
rootdir:	.byte	4	; FAT16: First Sector[=part1start+reservdsect+numfat*sectperfat], FAT32: First Cluster
sectperfat:	.byte	4	; Sectors per FAT as 32-bit integer
fat1start:	.byte	4	; First Sector of 1st FAT [=part1start+reservdsect]
datastart:	.byte	4	; First Sector of Data Section [=part1start+reservdsect+numfat*sectperfat+int((rootentries*32+bytespsect-1)/bytespsect)]
;
dirsectors:	.byte	1	; FAT16: Sectors in Root directory, FAT32: 0
numfat:		.byte	1	; Number of FAT entries
sectperclst:	.byte	1	; Sectors per Cluster, valid values are 1,2,4,16,32,64
VolStatus:	.byte	1	; Volume Status
.equ	MBRok	= 0		; MBR read successfully
.equ	VBRok	= 1		; VBR read successfully
.equ	FAT32	= 2		; Volume is FAT32
;
;--------------------------------------------------------------------------
;
;	Flash
;
	.cseg
.org	0
	rjmp	reset
.org	INT0addr
	rjmp	INTQ_isr
.org	INT1addr
	rjmp	INTI_isr
.org	PCI2addr
	rjmp	PC_INIT
.org	OC0Aaddr
	rjmp	OC0_isr
.org	URXC0addr
	rjmp	rxc0_isr
.org	UDRE0addr
	rjmp	dre0_isr
	
.org	INT_VECTORS_SIZE

;--------------------------------------------------------------------------
;
;	Interrupt driven serial input and output routines with a transmit and
;	a receive ring buffer. The ring buffers are page aligned and 256bytes
;	in other words by just using the low byte of the address we have an
;	automatic overflow. Each ring buffer has three variables associated.
;	An insert pointer, a remove pointer and a count. If count is zero the
;	ring buffer is empty, if count is 0xFF the ring buffer is full.
;	The transmitter uses the date register empty interrupt which is more
;	suitable for this scenario. In this way the serout routine only waits
;	for the transmit ring buffer not being full, then inserts the byte
;	increments the count and insert pointer and then activates the data
;	register empty interrupt.
;
;--------------------------------------------------------------------------
;
;	Receive Complete Interrupt
;
rxc0_isr:				;  7
	push	r8			;  2
	in	r8, SREG		;  1	Save status
	push	r16			;  2
	push	zl			;  2
	push	zh			;  2	Save pointer register
	lds	r16, UDR0		;  2	retreive the character
	lds	zl, rxcnt		;  2	test if buffer is not full
	inc	zl			;  1	else increment the level indicator
	breq	rxc0_overflow		;  1	receive buffer is full, discard
	sts	rxcnt, zl		;  2
	ldi	zh, high(rxring)	;  1	get the ring buffer insert pointer
	lds	zl, rxinptr		;  2
	st	Z+, r16			;  2	insert the character
	sts	rxinptr, zl		;  2	update insert pointer
rxc0_overflow:
	pop	zh			;  2
	pop	zl			;  2
	pop	r16			;  2
	out	SREG, r8		;  2
	pop	r8			;  2
	reti				;  2
;--------------------------------------------------------------------------
;
;	Data Register Empty Interrupt, i.e. another character may be sent to USART
;
dre0_isr:				;  7
	push	r8			;  2
	in	r8, SREG		;  1	if this interrupt is triggered we know
	push	r16			;  2
	push	zl			;  2	for sure that a character is waiting
	push	zh			;  2	
	lds	zl, txoutptr		;  2	Get remove pointer
	ldi	zh, high(txring)	;  1
	ld	r16, Z+			;  1	remove character
	sts	txoutptr, zl		;  2	update remove pointers
	sts	UDR0, r16		;  2	and copy the character to the UART
	lds	zl, txcnt		;  2	decrement the level indicator
	dec	zl			;  1
	sts	txcnt, zl		;  2
	brne	dre0_isr_next		;  1	Finished if there is more
	lds	zl, UCSR0B		;  2	when the ring buffer is now empty
	cbr	zl, (1<<UDRIE0)		;  1	we have sent the last character and
	sts	UCSR0B, zl		;  2	clear the data register empty interrupt
					;   	next call of serout will activate it again	
dre0_isr_next:
	pop	zh			;  2
	pop	zl			;  2
	pop	r16			;  2
	out	SREG, r8		;  1
	pop	r8			;  2
	reti				;  2
;--------------------------------------------------------------------------
;
;	Power Up
;
reset:
;
;	Waste time. Automount does not work if it is interrupted by the
;	second reset of the AVRISP MKII. 
;
	ldi	xl, low(0x8000)
	ldi	xh, high(0x8000)
	clr	r18
resetwait:
	inc	r18
	brne	resetwait
	adiw	xh:xl, 1
	brne	resetwait
;
;	Set Stack Pointer to top of RAM
;
	ldi	xl, low(RAMEND) 
	ldi	xh, high(RAMEND) 
	out	SPL, xl 
	out	SPH, xh
;
;	Init Constants
;
	clr	zero
;--------------------------------------------------------------------------
;
;	USART0
;	
;	Interactive user interface to control the AVR
;
;	115200baud, 8 data, no parity, 1 stop bit
;
;	Initialise Ring Buffer Pointers/Counters
;
	sts	txcnt, zero
	sts	txinptr, zero
	sts	txoutptr, zero
	sts	rxcnt, zero
	sts	rxinptr, zero
	sts	rxoutptr, zero
;
;	Set Baudrate
;
	ldi	r16, HIGH(UBRR0_VAL)
	ldi	r17, LOW(UBRR0_VAL)
	sts	UBRR0H, r16
	sts	UBRR0L, r17
;
;	Enable receiver and transmitter and receiver interrupt
;
	ldi	r16, (1<<RXEN0)|(1<<TXEN0)|(1<<RXCIE0)
	sts	UCSR0B, r16
;
;	Set frame format: 8data, 1stop bit, no parity
;
	ldi	r16, (3<<UCSZ00)
	sts	UCSR0C, r16
;--------------------------------------------------------------------------
;
;	Port A - Data Port
;
#define	dataportout	PORTA
#define dataportin	PINA
#define dataportdir	DDRA

	out	dataportdir, zero	; Initially configured as input

;--------------------------------------------------------------------------
;
;		Port B
;
;	Whenever we create the equates for the bits we also create defines that
;	can be used for sbi and cbi instructions and further defines for the ports
;	where these bits have been defined
;
;	b_xxxx	definition for sbi and cbi
;	i_xxxx	input port used for the bit, also used to toggle a bit
;	o_xxxx	output port used for the bit 
;
;
;	RLV12 Emulator Hardware Version 2 with new interface
;
.equ	WR		= 0		; Write Clock
.equ	CLK		= 1		; System Clock
.equ	CRDY		= 2		; Controller Ready
.equ	WP		= 3		; Write Protect input for SD-Card Cage
.equ	SPI_SS		= 4		; SPI
.equ	SPI_MOSI	= 5		; SPI
.equ	SPI_MISO	= 6		; SPI
.equ	SPI_SCK		= 7		; SPI

#define d_CRDY DDRB, CRDY
#define b_CRDY PORTB, CRDY
#define	b_WR PORTB, WR
#define t_WR PINB
#define	b_WP PORTB, WP
#define	i_WP PINB, WP
#define b_SPI_SS PORTB, SPI_SS

	ldi	r16, (1<<SPI_MOSI) | (1<<SPI_SS) | (1<<SPI_SCK) | (1<<WR) | (1<<CRDY)
	out	DDRB, r16
	cbi	b_WR
	sbi	b_CRDY
	sbi	b_WP			; Enable pull-up
;--------------------------------------------------------------------------
;
;		Port C
;
.equ	ABORT		= 0		; Abort DMA Cycle
.equ	ALER		= 1		; Address Latch Enable Register Read
.equ	DMG		= 2		; DMA Grant Status 
.equ	RD	 	= 3		; Read Data
.equ	INIT		= 4		; Q-Bus INIT
.equ	IACK		= 5		; Acknowledge Interrupt for CPLD
.equ	SIG	 	= 6		; Signal (SPARE1 on PCB)
.equ	DMR		= 7		; DMA Request

#define b_ABORT PORTC, ABORT
#define b_ALER PORTC, ALER
#define t_ALER PINC
#define i_DMG PINC, DMG
#define i_INIT PINC, INIT
#define	b_RD PORTC, RD
#define b_IACK PORTC, IACK
#define b_RA1 PORTC, RA1
#define b_SIG PORTC, SIG
#define	b_DMR PORTC, DMR

	out	PORTC, zero
	ldi	r16, (1<<ABORT) | (1<<ALER) | (1<<RD) | (1<<IACK) | (1<<SIG) | (1<<DMR)
	out	DDRC, r16
	sbi	b_ABORT			; Reset DMA Statemachine
	cbi	b_ABORT
	sbi	b_IACK			; Acknowledge interrupt
	cbi	b_IACK
	cbi	b_SIG
;--------------------------------------------------------------------------
;
;		Port D
;
.equ	RXD0		= 0		; AVR Console USART RXD
.equ	TXD0		= 1		; and TXD
.equ	INTQ		= 2		; Interrupt 0
.equ	INTI		= 3		; Interrupt 1
.equ	IRQST		= 4		; Interrupt Request
.equ	ALEW		= 5		; Address Latche Enabel Register Write
.equ	LED		= 6
.equ	CD		= 7

#define b_IRQST PORTD, IRQST
#define b_ALEW PORTD, ALEW
#define b_LED PORTD, LED
#define	b_CD PORTD, CD
#define	b_INTQ PORTD, INTQ
#define i_INTQ PIND, INTQ
#define b_INTI PORTD, INTI
#define i_INTI PIND, INTI
#define i_CD PIND, CD

	out	PORTD, zero
	sbi	b_CD			; Enable pull-up
	sbi	b_INTQ			; Enable pull-up
	sbi	b_INTI			; Enable pull-up
	ldi	r16, (1<<IRQST) | (1<<ALEW) | (1<<LED)
	out	DDRD, r16
;--------------------------------------------------------------------------
;
;	Initialise Device Registers
;
	ldi	zl, low(CSRL)
	ldi	zh, high(CSRL)
	ldi	r16, (1<<CSR_CRDY)	; Controller is ready
	std	Z+0, r16
	std	Z+1, zero
	std	Z+2, zero
	std	Z+3, zero
	std	Z+4, zero
	std	Z+5, zero
	std	Z+6, zero
	std	Z+7, zero
	std	Z+8, zero
	std	Z+9, zero
	std	Z+10, zero
	std	Z+11, zero
	std	Z+12, zero
	std	Z+13, zero
	std	Z+14, zero
	std	Z+15, zero
;--------------------------------------------------------------------------
;
;	Initialise heap
;
	ldi	xl, low(heapsize)
	ldi	xh, high(heapsize)
	ldi	yl, low(heapstart)
	ldi	yh, high(heapstart)
	
	sts	heap+0, zero		; Dummy record that starts the
	sts	heap+1, zero		; List of free blocks with size 0
	sts	heap+2, yl		; and points to the heap
	sts	heap+3, yh
	
	std	Y+0, xl			; Size of heap
	std	Y+1, xh
	std	Y+2, zero		; Just one block for now, no more
	std	Y+3, zero

;--------------------------------------------------------------------------
;
;	Soft ROM
;
	ldi	zl, low(2*rom173000)
	ldi	zh, high(2*rom173000)
	ldi	yl, low(softrombuffer)
	ldi	yh, high(softrombuffer)
	ldi	xl, low(softromsize)
	ldi	xh, high(softromsize)
softrom:
	lpm	r16, Z+
	st	Y+, r16
	sbiw	xh:xl, 1
	brne	softrom
;--------------------------------------------------------------------------
;
;	Initialise pcb queue
;
	sts	pcbqueue+0, zero
	sts	pcbqueue+1, zero
	sts	pcbcount, zero
	ldi	r16, 'C'
	sts	pcbdisk, r16
;--------------------------------------------------------------------------
;
;	Initialise data structures
;
	ldi	yl, low(pfatcluster)
	ldi	yh, high(pfatcluster)
	std	Y+P_Cluster+0, zero
	std	Y+P_Cluster+1, zero
	std	Y+P_Cluster+2, zero
	std	Y+P_Cluster+3, zero
	std	Y+P_Sector+0, zero
	std	Y+P_Sector+1, zero
	std	Y+P_Sector+2, zero
	std	Y+P_Sector+3, zero
	ldi	r16, low(pfatbuffer)
	ldi	r17, high(pfatbuffer)
	std	Y+P_Address+0, r16
	std	Y+P_Address+1, r17
	sts	NameBuffer, zero	; No Name in NameBuffer
	ldi	r16, low(plfilename)
	ldi	r17, high(plfilename)
	sts	pdircluster+P_LongFileN+0, r16
	sts	pdircluster+P_LongFileN+1, r17
	ldi	r18, units * unitsize
	ldi	yl, low(unittable)
	ldi	yh, high(unittable)
resetunittable:
	st	Y+, zero
	dec	r18
	brne	resetunittable
	sts	VolStatus, zero		; Invalidate FAT partition
	sts	sd_status, zero		; Invalidate SD-Card
;--------------------------------------------------------------------------
;
;	Adjust the existing log_pointer to a valid value, that is it must
;	point to the log_buffer and must be a multiple of 8.
;
	lds	xl, log_pointer+0
	lds	xh, log_pointer+1
	andi	xl, 0xF8
	andi	xh, high(log_buffersize-1)
	ori	xh, high(log_buffer)
	sts	log_pointer+0, xl
	sts	log_pointer+1, xh
	ldi	r16, log__units | (1<<log__reg) | (1<<log__iack) | (1<<log__pbn)
	out	GPIOR0, r16		; Initialise logging flags
	sts	log_message, zero	; invalidate pending message
	rcall	splashscreen
;--------------------------------------------------------------------------
;
;	Initialise Autoboot Flag
;
	sts	autoboot, zero
;--------------------------------------------------------------------------
;
;	External Interrupt
;
;	ISCn1	ISCn0
;	0	0	The low level of INTn generates an interrupt request
;	0	1	Any edge of INTn generates asynchronously an interrupt request
;	1	0	The falling edge of INTn generates asynchronously an interrupt request
;	1	1	The rising edge of INTn generates asynchronously an interrupt request
;
;	INT0	INTQ	DATI/DATO
;	INT1	INTI	IACK
;
;	We use low level to trigger the interrupt. It is the task of the isr to
;	acknowledge the interrupt of the CPLD by generating a short pulse on
;	IACK (PORT C).
;
	sts	INTQ_count, zero
	sts	INTI_count, zero
	sts	INIT_count, zero
	cbi	b_IACK			; Is initially set!
	ldi	r16, (0<<ISC11) | (0<<ISC10) | (0<<ISC01) | (0<<ISC00)
	sts	EICRA, r16		; Set interrupt type
	ldi	r16, (1<<INT1) | (1<<INT0)
	out	EIMSK, r16		; Enable Interrupts
;--------------------------------------------------------------------------
;
;	Pin Change INIT at PORTC
;
	sts	INIT_count, zero
	ldi	r16, (1<<INIT)
	sts	PCMSK2, r16
	ldi	r16, (1<<PCIE2)
	sts	PCICR, r16
;--------------------------------------------------------------------------
;
;	Start timer 2 as free running clock counting "ticks"
;
;	We are running at 22MHz and we can assume that a PDP-11 is not 
;	faster than accessing the RLV12 every 5usec, so we can use
;	a prescaler of 128, which gives a resolution of 5.78usec
;
;	CS22	CS21	CS20
;	0	0	0	Stop
;	0	0	1	1
;	0	1	0	8
;	0	1	1	32
;	1	0	0	64
;	1	0	1	128	<---
;	1	1	0	256
;	1	1	1	1024
;
	sts	TCNT2, zero
	sts	TCCR2A, zero
	ldi	r16, (1<<CS22) | (0<<CS21) | (1<<CS20)
	sts	TCCR2B, r16
;--------------------------------------------------------------------------
;
;	Start the debounce timer (5msec) and initialise SD-Card status vars.
;
	ldi	r16, 0xff
	sts	s_CDINS, r16
	sts	s_CDINSD, zero
	sts	s_CDINSC, zero
	sts	s_CDREM, r16
	sts	s_CDREMD, zero
	sts	s_CDREMC, zero
	out	TCNT0, zero
	ldi	r16, (1<<WGM01)
	out	TCCR0A, r16
	ldi	r16, (1<<CS02) | (0<<CS01) | (1<<CS00)	; Prescaler 1024
	out	TCCR0B, r16
	ldi	r16, int(F_CPU/1024*0.005)
	out	OCR0A, r16
	ldi	r16, (1<<OCIE0A)
	sts	TIMSK0, r16
	
;--------------------------------------------------------------------------
;
;	Disable Boot ROM
;
	cbi	b_RD
	ldi	r16, 0xff
	out	dataportdir, r16
	ldi	r16, 7
	out	dataportout, r16
	sbi	b_ALEW
	cbi	b_ALEW
	out	dataportout, zero	; Disable Boot ROM
	sts	romstatus, zero		;
;	out	dataportout, one	; Enable Boot ROM
;	sts	romstatus, one		; 
	sbi	b_WR
	cbi	b_WR
;==========================================================================
;
;	Enable interrutps, enter CLI command line interface, start emulator
;
	sei
readcmd:
	sts	InputBuffer, zero
	sts	HistoryBuffer, zero

readcmd010:
	ldi	r22, ']'		; Prompt
	ldi	r24, low(InputBuffer)
	ldi	r25, high(InputBuffer)
	call	readcmdline
	cpi	r24, 0;cmdok		; 
	brne	readcmd040
	lds	r18, InputBuffer
	cpi	r18, 0
	breq	readcmd010
	cpi	r18, CR
	breq	readcmd010		; do not save empty command
	ldi	xl, low(InputBuffer)	; Command
	ldi	xh, high(InputBuffer)
	ldi	zl, low(HistoryBuffer)
	ldi	zh, high(HistoryBuffer)
readcmd020:
	ld	r18, X+
	st	Z+, r18
	cpi	r18, 0
	breq	readcmd030
	cpi	r18, CR
	breq	readcmd030
	rjmp	readcmd020
readcmd030:
	ldi	r18, CR
	st	-X, r18			; Make sure buffer is CR terminated
	call	seroutcrlf
	ldi	xl, low(InputBuffer)	; Command
	ldi	xh, high(InputBuffer)
	ldi	zl, low(2*commandlist)	; Parser table
	ldi	zh, high(2*commandlist)
	sts	tpflags, zero
	call	scancommand		; call TParse
	sts	InputBuffer, zero
	rjmp	readcmd010

readcmd040:
	cpi	r24, 1;cmdup		; 
	brne	readcmd060
	ldi	xl, low(HistoryBuffer)
	ldi	xh, high(HistoryBuffer)
	ldi	zl, low(InputBuffer)	; Command
	ldi	zh, high(InputBuffer)
readcmd050:
	ld	r18, X+
	st	Z+, r18
	cpi	r18, 0
	breq	readcmd010
	cpi	r18, CR
	breq	readcmd010
	rjmp	readcmd050

readcmd060:
	cpi	r24, 2;cmddown
	ldi	xl, low(InputBuffer)	; Command
	ldi	xh, high(InputBuffer)
	ldi	r24, SPACE
	clr	r25
readcmd070:				; Init the buffer
	st	X+, r24
	inc	r25
	brpl	readcmd070
	sts	InputBuffer, zero
	rjmp	readcmd010
;--------------------------------------------------------------------------
;
;	Q-Bus Interrupt handlers
;
;--------------------------------------------------------------------------
.include "Qbridge11-rlv12-qbus-v2.asm"

;--------------------------------------------------------------------------
;
;	Clock Interrupt for recurring tasks
;
;	-	Detect SD-Card insertion and removal
;	-	Activity LED shutoff
;
OC0_isr:
	push	r8
	in	r8, SREG		; we can use the  interrupt registers
	push	r16
	clc				;
	sbic	i_CD			;
	sec				;
	lds	r16, s_CDINS		;
	adc	r16, r16		;
	ori	r16, 0xe0
	sts	s_CDINS, r16
	cpi	r16, 0xf0
	brne	OC0_010
	ldi	r16, 0xff
	sts	s_CDINSD, r16
;	lds	r16, s_CDINSC
;	inc	r16
;	sts	s_CDINSC, r16
OC0_010:
	clc
	sbis	i_CD
	sec
	lds	r16, s_CDREM
	adc	r16, r16
	ori	r16, 0xe0
	sts	s_CDREM, r16
	cpi	r16, 0xf0
	brne	OC0_020
	ldi	r16, 0xff
	sts	s_CDREMD, r16
;	lds	r16, s_CDREMC
;	inc	r16
;	sts	s_CDREMC, r16
OC0_020:
	lds	r16, led_oneshot
	dec	r16
	brpl	OC0_030
	cbi	b_LED
	clr	r16
OC0_030:
	sts	led_oneshot, r16
	pop	r16
	out	SREG, r8
	pop	r8
	reti
;--------------------------------------------------------------------------
;
;	SD-Card insertion and removal handling
;
;	To detect insertion and removal we use a debounce routine found in the 
;	internet which uses a timer interrupt and a shifted value to detect a 
; 	raising or a falling edge on CD input. When a edge has been detected
;	either s_CDINSD or s_CDREMD are set to 0xFF in the timer interrupt.
;
;	sdcardinsert and sdcardremove are called by the dispatcher when one of
;	the corresponding variables is set. The action routines are then 
;	responsible to reset the flag (i.e. store 0x00 into the variables)
;

;-----------------------------------------------------------------------------
;
;	The scheduler is linked to 'serin'. As serin is supposed to preserve
;	all registers we must make sure that all routines scheduled within
;	'serin' must preserve all registers as well. The only register that
;	may be used to return a status is r24, as this is also the register
;	that is used by 'serin' to return the character. The following
;	actions may be called by 'serin'. This of course applies as well
;	to all sub-routines called by the actions.
;
;	sdcardinsert
;	sdcardremove
;	rlv12_autoboot
;	rlv12_sub
;
;-----------------------------------------------------------------------------
;
;	sdcardinsert
;
sdcardinsert:
	push	r16			; Temporary Register
	push	r18			; Counter
	push	r24
	push	r25
	push	yl			; Parameter Block
	push	yh			;
	push	zl			; Serial Buffer Pointer
	push	zh			;
	sts	s_CDINSD, zero
	call	print
	.db	LF, CR, "SD Card inserted -> automount", LF, CR, 0x00
	rcall	sdcardsetspi
	brcc	sdcardinsert010
	rjmp	sdcardinsert900
	
sdcardinsert010:
	rcall	sdcardifc
	brcc	sdcardinsert020
	rjmp	sdcardinsert900
	
sdcardinsert020:
	rcall	sdcardinit
	brcc	sdcardinsert030
	rjmp	sdcardinsert900
	
sdcardinsert030:
	rcall	sdcardreadocr
	brcc	sdcardinsert040
	rjmp	sdcardinsert900

sdcardinsert040:
	lds	r24, sd_status
	sts	pprint+0, r24
	sbrc	r24, sd__ccs_bp
	rjmp	sdcardinsert050		; Drive Capacity set non need to set blklen
	rcall	sdcardblklen
	brcc	sdcardinsert050
	rjmp	sdcardinsert900

sdcardinsert050:
	call	print
	.db	"Mounting SD-Card!", CR, LF, 0
	call	MountVolume
	sts	pprint+12, r24
	sts	pprint+13, zero
	brcc	sdcardvolumeok
	call	print
	.db	"IO Error:", 0x8a, ", ", 0x8b, " Error ID:", 0xcc, CR, LF, 0, 0
	rjmp	sdcardinsert900

sdcardvolumeok:
	call	print
	.db	"Mount Volume Success!", CR ,LF, 0

sdcardinsert900:
	ldi	r16, CTRL_R
	cli				;;;
	lds	zl, rxcnt		;;;	test if buffer is not full
	inc	zl			;;;
	breq	sdcardinsertov		;;;	receive buffer is full, discard
	sts	rxcnt, zl		;;;
	ldi	zh, high(rxring)	;;;	get the ring buffer insert pointer
	lds	zl, rxinptr		;;;
	st	Z+, r16			;;;	insert the <CR>
	sts	rxinptr, zl		;;;	update insert pointer
sdcardinsertov:				;;;
	sei				;;;
	pop	zh			;
	pop	zl			;
	pop	yh
	pop	yl
	pop	r25
	pop	r24
	pop	r18
	pop	r16
	ret
	
sdcardsetspi:				; Monitor wrapper
	call	SD_CARD_SPI
	sts	pprint+0, r24		; Save Return status
	brcs	sdcardsetspierr
	call	print
	.db	"Set SPI Mode Succeeded Counts:", 0x80, CR, LF, 0x00
	clc
	ret
sdcardsetspierr:
	call	print
	.db	"Set SPI Mode failed Error 0x", 0x80, CR, LF, 0x00
	sec
	ret
;
;
;
sdcardifc:
	call	SD_CARD_IFC
	brcs	sdcardifcerr
	rcall	sdcardcopyrsp
	call	print
	.db	"Interface Condition Read: 0x", 0x80, 0x81, 0x82, 0x83, 0x84, LF, CR, 0
	clc
	ret
sdcardifcerr:
	call	print
	.db	"Failed requesting IFC Error: ", CR, LF, 0
	rcall	sdcardprintr1
	sec
	ret
;
;
;
sdcardinit:
	call	SD_CARD_INIT
	brcs	sdcardiniterr
	lds	r16, sd_rsp+0
	sts	pprint+0, r16
	sts	pprint+1, r24		; Save Retries
	call	print
	.db	"Initialise Card RC:0x", 0x80, ", # of retries: 0x", 0x81, LF, CR, 0
	clc
	ret
sdcardiniterr:
	call	print
	.db	"Initialise Card Failed Error:", LF, CR, 0x00
	lds	r24, sd_rsp+0
	rcall	sdcardprintr1
	sec
	ret
;
;	Read Operational Condition Register
;
sdcardreadocr:
	call	SD_CARD_READOCR
	rcall	sdcardcopyrsp
	call	print
	.db	"Operation Condition Reg: 0x", 0x80, 0x81, 0x82, 0x83, 0x84, 0x85, LF, CR, 0
	lds	r16, sd_status
	sbrs	r16, sd__ccs_bp
	rjmp	sdcardreadocr010
	call	print
	.db	TAB, "High Capacity (CCS) ", CR, LF, 0
sdcardreadocr010:
	sbrs	r16, sd__reset_bp
	rjmp	sdcardreadocr020
	call	print
	.db	TAB, "Card successfully reset ", CR, LF, 0
sdcardreadocr020:
	sbrs	r16, sd__init_bp
	rjmp	sdcardreadocr030
	call	print
	.db	TAB, "Card successfully initialized ", CR, LF, 0
sdcardreadocr030:
	sbrs	r16, sd__ver2_bp
	rjmp	sdcardreadocr040
	call	print
	.db	TAB, "Version 2 Card", CR, LF, 0
sdcardreadocr040:
	clc
	ret
;
;
;
;
;	Set Block Length
;
sdcardblklen:
	ldi	yl, low(pcluster)	; Setup parameter block for general IO
	ldi	yh, high(pcluster)	; 
	call	SD_CARD_BLKLEN
	brcs	sdcardblklenerr
	ldd	r16, Y+P_IOStatus
	sts	pprint+0, r16
	call	print
	.db	"Set Block Length RC:", 0x80, LF, CR, 0x00
	clc
	ret
sdcardblklenerr:
	call	print
	.db	"Failed setting block length. ", CR, LF, 0x00
	sec
	ret
;
;
;
sdcardcopyrsp:
	ldi	yl, low(sd_rsp)
	ldi	yh, high(sd_rsp)
	ldi	zl, low(pprint)
	ldi	zh, high(pprint)
	ld	r24, Y+
	st	Z+, r24
	ld	r24, Y+
	st	Z+, r24
	ld	r24, Y+
	st	Z+, r24
	ld	r24, Y+
	st	Z+, r24
	ld	r24, Y+
	st	Z+, r24
	ld	r24, Y+
	st	Z+, r24
	ret
;
;
;
sdcardprintr1:
	sbrs	r24, 7
	rjmp	sdcardifcerr010
	call	print
	.db	TAB, "Error MSB = 1 ", CR, LF, 0
	ret
sdcardifcerr010:
	sbrs	r24, 6
	rjmp	sdcardifcerr020
	call	print
	.db	TAB, "Parameter Error ", CR, LF, 0
	ret
sdcardifcerr020:
	sbrs	r24, 5
	rjmp	sdcardifcerr030
	call	print
	.db	TAB, "Address Error ", CR, LF, 0
sdcardifcerr030:
	sbrs	r24, 4
	rjmp	sdcardifcerr040
	call	print
	.db	TAB, "Erase Sequence Error", CR, LF, 0
sdcardifcerr040:
	sbrs	r24, 3
	rjmp	sdcardifcerr050
	call	print
	.db	TAB, "CRC Error ", CR, LF, 0
sdcardifcerr050:
	sbrs	r24, 1
	rjmp	sdcardifcerr060
	call	print
	.db	TAB, "Erase Reset Error ", CR, LF, 0
sdcardifcerr060:
	ret
;--------------------------------------------------------------------------
;
;
;
sdcardremove:
	push	r16
	push	r24
	push	r25
	push	yl
	push	yh
	push	zl
	push	zh
	sts	s_CDREMD, zero
;
;	Reset partition control block queue
;
	ldi	yl, low(pcbqueue)
	ldi	yh, high(pcbqueue)
	sts	pcbqueue+0, zero
	sts	pcbqueue+1, zero
	sts	pcbcount, zero
	ldi	r16, 'C'
	sts	pcbdisk, r16
	ldd	zl, Y+0
	ldd	zh, Y+1			; Get next packet

sdcardremove020:	
	sbiw	zh:zl, 0	
	breq	sdcardremove030		; eq then done
	movw	r25:r24, zh:zl		; packet to free
	movw	yh:yl, zh:zl		; copy pointer
	ldd	zl, Y+0
	ldd	zh, Y+1			; Get next packet
	call	free			; Free this packet
	rjmp	sdcardremove020	

sdcardremove030:
	sts	VolStatus, zero		; Invalidate FAT partition
	sts	sd_status, zero		; Invalidate SD-Card
	call	rlv12_init
	call	print
	.db	LF, CR, "SD Card removed", LF, CR, 0x00
	
	ldi	r16, CTRL_R
	cli
	lds	zl, rxcnt		;;;	test if buffer is not full
	inc	zl			;;;
	breq	sdcardremoveov		;;;	receive buffer is full, discard
	sts	rxcnt, zl		;;;
	ldi	zh, high(rxring)	;;;	get the ring buffer insert pointer
	lds	zl, rxinptr		;;;
	st	Z+, r16			;;;	insert the <CR>
	sts	rxinptr, zl		;;;	update insert pointer
sdcardremoveov:

	pop	zh
	pop	zl
	pop	yh
	pop	yl
	pop	r25
	pop	r24
	pop	r16
	ret
;--------------------------------------------------------------------------
;
;
;
seroutcrlf:
	ldi	r24, CR
	rcall	serout
	ldi	r24, LF

serout:
;	rjmp	seroutold
	push	zl
	push	zh

	lds	zl, txcnt		; Get number of characters in buffer
	inc	zl			; Is buffer full?
	brne	seroutnowait		; Not full, so we will not wait
seroutwaitb:
	lds	zl, txcnt		; Get number of characters in buffer
	inc	zl			; Is buffer full?
	breq	seroutwaitb		; Then wait until there is room
seroutnowait:
	ldi	zh, high(txring)
	cli				;;;

	lds	zl, txinptr		;;; Get input pointer
	st	Z+, r24			;;; Save Character
	sts	txinptr, zl		;;; Update input pinter

	lds	zl, txcnt		;;; Increment count of characters
	inc	zl			;;;
	sts	txcnt, zl		;;;
	lds	zl, UCSR0B		;;; Activate the data register empty
	sbr	zl, (1<<UDRIE0)		;;; interrupt so the ISR picks up the
	sts	UCSR0B, zl		;;; queued character(s)
	sei				;;; 

	pop	zh
	pop	zl
	ret

seroutold:
	push	r24
seroutwait:
	lds	r24, UCSR0A
	sbrs	r24, UDRE0
	rjmp	seroutwait
	pop	r24
	sts	UDR0,r24
	ret			
;--------------------------------------------------------------------------
;
;	The task dispatcher is currently linked to the serin routine. 
;
serin:
	lds	r24, rxcnt
	tst	r24
	brne	serin010
	lds	r24, s_CDINSD
	sbrc	r24, 0
	rcall	sdcardinsert		; SD-Card was inserted
	lds	r24, s_CDREMD
	sbrc	r24, 0
	rcall	sdcardremove		; SD-Card was removed
	sbis	b_CRDY			; Has a command been scheduled?
	call	rlv12_sub		; Call Command Execution Sub
	lds	r24, autoboot
	cpse	r24, zero
	call	rlv12_autoboot
	rjmp	serin
	
serin010:
	push	zl
	push	zh
	ldi	zh, high(rxring)
	cli				;;;
	lds	zl, rxoutptr		;;;
	ld	r24, Z+			;;;
	sts	rxoutptr, zl		;;;
	lds	zl, rxcnt		;;;
	dec	zl			;;;
	sts	rxcnt, zl		;;;
	sei				;;;
	pop	zh
	pop	zl
	ret
;--------------------------------------------------------------------------
;
;	Splashscreen
;
splashscreen:
	ldi	zl, low(2*hello)
	ldi	zh, high(2*hello)

splashsend:
	lpm	r24 ,Z+ 
	tst	r24
	breq	splashdone
	rcall	serout
	rjmp	splashsend

splashdone:
	ret

load:
	ld	r17, X
	cpi	xh, 0x80
	brlo	loaduser010
	push	zl
	push	zh
	movw	Z, X
	andi	zh, 0x7F
;	subi	zl, low(-2*romtest)
;	sbci	zh, high(-2*romtest)
	lpm	r17, Z
	pop	zh
	pop	zl
loaduser010:
	ret
;
;	Monitor and monitor action routines
;
.include "DMA-macros.inc"
.include "Qbridge11-monitor.asm"
;
;	Formatted print output
;
.include "../include/print-v2-0.inc"
	.cseg
.include "../include/print-v2-1.asm"
;
;	Table driven parser
;
.include "../include/tparse-v2-0.inc"
	.cseg
.include "../include/tparse-v2-1.asm"
;
;	malloc() and free()
;
.include "../include/malloc-v1-0.asm"
;
;	Read Command Line Subroutine
;
.include "readcmdline.asm"
;
;	SD-Card IO routines
;
.include "SD-Card-IO-v2-1.asm"
;
;	Include the rlv12 emulator
;
.include "Qbridge11-rlv12-v2.asm"
;
;	Include some of the application specific action routines related to
;	the SD-Card interface or the FAT-16/32 library
;
.include "Qbridge11-cli.inc"
.include "Qbridge11-cli-action.asm"
.include "Qbridge11-cli.asm"
.include "Qbridge11-help.asm"
.include "Qbridge11-readinit.asm"
.include "Qbridge11-fdisk.asm"
.include "FAT-library/FAT-library-v1-3.asm"	
.include "Qbridge11-mountvolume-v1-0.asm"


;
;	Include the more application specific action routines which
;	interact with the PDP-11 or are related to the rvl12 emulator
;
.include "Qbridge11-logging.asm"

;--------------------------------------------------------------------------
;


hello:	.db	CR, LF, CR, LF
;		"----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8"
#ifdef cpld
	.db	"Qbridge RLV12 Emulator V2 Vers 0.9.6   ",__DATE__ ,"  CPLD Interface ", '0'+rlv12/10, '0'+rlv12%10, " active  ", CR, LF
#else
	.db	"Qbridge RLV12 Emulator V2 Vers 0.9.6   ",__DATE__ ,"  CPLD Interface disabled   ", CR, LF
#endif
	.dw	0




