;--------------------------------------------------------------------------
;
;	To-do-list
;
;	-	Conditional assembly for various processor types
;	-	Define a dedicated register for the SPI primitives to exchange the
;		data with the caller
;	-	rename SD_CARD_CMD55 to SD_SEND_CMD55
;
;	2022-04-16
;		Clean-Up and use new parameter block fields P_IOStatus and	P_Duration
;		All Functions must be called with Y pointing to a IO Block
;
;	2022-04-19
;		Further clean-up, remove conditional, we only will use this
;		for the Atmega1284P.
;
;	SD-Card IO Routines	
;
;	I.	Credits
;		The document which for me inlcuded the most useful information is
;		a PDF by Simeon Maxein.
;
;	1.	SPI Primitives
;		These routines transfere a single byte from and to the SPI device
;		and are used when performance is not impartant, like initialisation
;		command phase and read status. 
;	1.1	SPI_WRITE_DUMMY
;		Writes 0xff to the device. Typically used to get a byte from the
;		SPI device
;	1.2	SPI_WRITE_BYTE
;		Write a byte
;	1.3 	SPI_CLEAR
;		Deselect the device. First a dummy byte is sent to flush any 
;		outstanding response. The response is discarded. Then the device
;		is deselected and another dummy byte is written to make sure
;		we end with MOSI in high state
;	1.4	SPI_GET_RESPONSE
;		This routine is used to get a command response byte. This is
;		anything but 0xff. Also this routine only waits for a certain
;		amount of time. When a time out occurs it will return 0xff else
;		it will return the response.
;
;	2.	SD Card Basic Command Routines
;		These routines write a 6 byte command to the SD Card
;
;	2.1	SD_SEND_CMD
;	2.2	SD_CARD_CMD55
;
;--------------------------------------------------------------------------
;
;	I	.dw	mon_sd_card_spi		SD_CARD_SPI
;	H	.dw	mon_sd_card_ifc		SD_CARD_IFC
;	J	.dw	mon_sd_card_init	SD_CARD_INIT
;	K	.dw	mon_sd_card_readocr	SD_CARD_READOCR
;	L	.dw	mon_sd_card_blklen	SD_CARD_BLKLEN
;
;--------------------------------------------------------------------------
;
;	SPI Primitives, these primitives destroy r24!
;
SPI_WRITE_DUMMY:			; Write a dummy 0xFF
	ldi	r24, 0xFF
SPI_WRITE_BYTE:				; Alternate Write Entry Point
	out	SPDR, r24		; to write what is set in 'r24'
SPI_WRITE_BYTE_L:
	in	r24, SPSR
	sbrs	r24, SPIF		; Poll SPI End of Transmission Flag
	rjmp	SPI_WRITE_BYTE_L
	ret				; Done
	
SPI_CLEAR:
	rcall	SPI_WRITE_DUMMY		; Flush any remaining byte
	sbi	b_SPI_SS		; Deselect SD-Card
	rcall	SPI_WRITE_DUMMY		; Make sure we end in MOSI=high
	ret

;
;	Get first or only (R1) response from the SD-Card
;	  7                           0
;	+---+---+---+---+---+---+---+---+
;	| 0 |   |   |   |   |   |   |   |
;	+---+---+---+---+---+---+---+---+
;	      |   |   |   |   |   |   +--- in idle state
;	      |   |   |   |   |   +------- erase reset
;	      |   |   |   |   +------------illegal command
;	      |   |   |   +----------------com crc error
;	      |   |   +--------------------erase sequence error
;	      |   +------------------------address error
;	      +----------------------------parameter error
;
SPI_GET_RESPONSE:			; 
	push	r18
	ldi	r18, -sdspi_o
SPI_GET_RES_010:
	rcall	SPI_WRITE_DUMMY
	in	r24, SPDR
	cpi	r24, 0xff
	brne	SPI_GET_RES_020
	inc	r18
	brne	SPI_GET_RES_010
SPI_GET_RES_020:
	pop	r18
	ret
;--------------------------------------------------------------------------
;
;	SD card basic command routines
;
SD_SEND_CMD:				; Send a 6 byte command
	sbi	b_SPI_SS		; Deselect SD-Card
	nop
	nop
	nop
	nop
	rcall	SPI_WRITE_DUMMY		; Set SPI to known state
	cbi	b_SPI_SS		; Select SD-Card
	lds	r24, sd_cmd+0		; Send 6 command bytes
	rcall	SPI_WRITE_BYTE
	lds	r24, sd_cmd+1
	rcall	SPI_WRITE_BYTE
	lds	r24, sd_cmd+2
	rcall	SPI_WRITE_BYTE
	lds	r24, sd_cmd+3
	rcall	SPI_WRITE_BYTE
	lds	r24, sd_cmd+4
	rcall	SPI_WRITE_BYTE
	lds	r24, sd_cmd+5
	rcall	SPI_WRITE_BYTE
	ret				; Return with SD-Card selected
;
;	SD CMD55 Prefix for ACMDs
;
SD_CARD_CMD55:
	sbi	b_SPI_SS		; Deselect SD-Card
	nop
	nop
	nop
	nop
	rcall	SPI_WRITE_DUMMY		; Set SPI to known state
	cbi	b_SPI_SS		; Select SD-Card
	ldi 	r24, 0x77
	sts	sd_cmd+0, r24
	sts	sd_cmd+1, zero
	sts	sd_cmd+2, zero
	sts	sd_cmd+3, zero
	sts	sd_cmd+4, zero
	ldi	r24, 1
	sts	sd_cmd+5, r24
	rcall	SD_SEND_CMD		; Send Command
	rcall	SPI_GET_RESPONSE
	sts	sd_rsp+0, r24
	rcall	SPI_CLEAR
	ret
;--------------------------------------------------------------------------
;
;	SD-card primitives
;
;	General interface conventions
;
;	-	On success the carry is cleared
;	-	If an error occurs the carry is set
;	-	P_IOStatus has the return code
;
;	SD_CARD_SPI	(CMD0)
;	Descr.		Set the SD-Card in SPI Mode
;	Input		none
;	Output		CC	Card response was success
;			CS	Card command error
;
;	sd_status	sd_reset	set when card responded successfully
;
;	SD_CARD_IFC	(CMD8)
;	Descr.		Request the interface condition. This command is only
;			supported by Version 2 SD-Cards. To detect Version 2
;			cards you must send this command, Version 1 cards will
;			answer with "invalid command".
;	Input		none
;	Output		CC	Card response was success
;			CS	Card reported error
;
;	sd_status	sd_ver2		set when command was accepted. Only version 2
;					cards accept this command
;					when set we also set HCS (Host Capacity Support)
;					during initialisation
;
;	SD_CARD_READOCR	(CMD58)
;	Descr.		Reqeust Operating Conditions. This command is only
;			supported by Version 2 SD-Cards. When the initialisation
;			is finished the card will set CCS in the response if it
;			is a HC/XC card. You should call this function when SD_CARD_IFC
;			was called successfully
;	Input		none
;	Output		CC	Card response was successful
;			CS	Card reported error
;
;	sd_status	sd_ccs	if card reports HCS then we set this bit in the status
;				note that the mask of sd_ccs (0x40) is equivalent to 
;				bit30 in the SD-Card OCR, i.e. bit6 in the first byte.
;				Only cards with >2Gbyte will normally report CCS. Even
;				Version 2 cards with 2Gbyte or less report no CCS.
;				When set SD_CARD_XLT will use block address else it
;				will translate the block address to a byte address.
;
;	SD_CARD_INIT	(ACMD41)
;	Descr.		Requests the card to initialise. Note ACMD41 is only
;			recognised by SD-Cards but not by MMC cards.
;	Input		none
;	Output		CC	Initialisation successfull P_IOStatus+1 contains the number of retries
;			CS	Initialisation unsuccessfull P_IOStatus contains last response
;
;	sd_status	sd_init	set if card could be initialised successfully
;
;	SD_CARD_BLKLEN	(CDM16)
;	Descr.		Set the block length, this is only required for standard
;			cards that do not support card capacity, this is SD-cards
;			with 2Gbyte or less capacity.
;	Input		none
;	Output		piostatus contains the response
;
;	Initialisation procedure
;
;	SD_CARD_SPI				CMD0
;	SD_CARD_IFC		-> V2?		CMD8
;	SD_CARD_INIT				ACMD41
;	if V2 SD_CARD_READOCR			CMD58
;	if not HCXC SD_CARD_BLKLEN		CMD16
;
;--------------------------------------------------------------------------
;	1) SD-Card Set SPI Mode				Monitor Command: I
;
;	After Power-Up the SD-Card is in SD mode, however it can be switched to
;	SPI mode by sending CMD0 after a power-on reset. Note that you cannot
;	exit SPI mode and once the card has been initialised it will no longer
;	recognize initialisation commands. Therefore you must remove and insert
;	the SD-Card to be able to reinitialize the card again.
;
;
SD_CARD_SPI:
	sts	sd_status, zero
	sbi	b_SPI_SS		; Deselect SD-Card
	ldi	r18, 10
SD_CARD_W1:
	dec	r18			; Wait some time
	brne	SD_CARD_W1

	ldi	r24, (0<<SPIE)|(1<<SPE)|(0<<DORD)|(1<<MSTR)|(0<<CPOL)|(0<<CPHA)|(1<<SPR1)|(1<<SPR0)
	out	SPCR, r24
	ldi	r24, (0<<SPI2X)	; Initialise uses slow clock
	out	SPSR, r24		; Initialise SPI interface
	ldi	r18, 15
SD_CARD_L1:
	rcall	SPI_WRITE_DUMMY		; Create >74 clock cycles with SD-Card deselected
	dec	r18
	brne	SD_CARD_L1
;
;	SD-Card Software Reset
;
	ldi	r24, 0x40		; Setup CMD0 = Reset
	sts	sd_cmd+0, r24
	sts	sd_cmd+1, zero
	sts	sd_cmd+2, zero
	sts	sd_cmd+3, zero
	sts	sd_cmd+4, zero
	ldi	r24, 0x95		; With precalculated CRC
	sts	sd_cmd+5, r24
	rcall	SD_SEND_CMD		; Send Command
;
	ldi	r18, -sdspi_i		; Inner time-out
SD_CARD_L2:
	rcall	SPI_WRITE_DUMMY		; Send dummy byte
	in	r24, SPDR
	cpi	r24, 0xff		; Did SD-Card answer, 0xff means no answer
	brne	SD_CARD_DONE		; Yes
	inc	r18			; Try one more time
	brne	SD_CARD_L2		; yes
;
SD_CARD_DONE:
	cpi	r24, 0x01		; Is Answer = Success (now in idle state)
	breq	SD_CARD_SUCCESS		; Yes
	rcall	SPI_CLEAR
	sec				; Error
	ret

SD_CARD_SUCCESS:	
	subi	r18, -sdspi_i
	push	r18
	rcall	SPI_CLEAR
	ldi	r24, sd__reset_bm
	sts	sd_status, r24		; Mark status
	pop	r24			;
	clc				; Success
	ret
;
;	2) SD-Card Interface Condition	Monitor Command: H
;
;	After CMD0 has been sent successfully and the card is in idle state
;	the interface conditions should be requested by the host using CMD8
;
;	CMD8 is only recognized by cards that conform to the SD specifications
;	version 2. Version 1 cards will return "invalid command" in the status
;	byte.
;
SD_CARD_IFC:
	ldi	r24, 0x48		; CMD8 Send Interface Condition
	sts	sd_cmd+0, r24
	sts	sd_cmd+1, zero
	sts	sd_cmd+2, zero
	ldi	r24, 1
	sts	sd_cmd+3, r24		; 2.7-3.6V is applied to SD Card
	ldi	r24, 0xAA
	sts	sd_cmd+4, r24		; Check Pattern that will be returned in response
	ldi	r24, 0x87
	sts	sd_cmd+5, r24		; Checksum
	rcall	SD_SEND_CMD
	rcall	SPI_GET_RESPONSE
	sts	sd_rsp+0, r24		; Status
	cpi	r24, 0xff		; Did we get a response?
	breq	SD_CARD_IF_ERR		; Error occured
	sbrc	r24, 2			; Illegal Command
	rjmp	SD_CARD_IF_ILL		; This is either a Version 1 Card or a MMC 
	cpi	r24, 1			; In Idle State, aka command accepted
	brne	SD_CARD_IF_ERR		; Else we have another error
	rcall	SPI_WRITE_DUMMY		; get the rest of the R7 response which is
	in	r24, SPDR		; 5 bytes long
	sts	sd_rsp+1, r24
	rcall	SPI_WRITE_DUMMY
	in	r24, SPDR
	sts	sd_rsp+2, r24
	rcall	SPI_WRITE_DUMMY
	in	r24, SPDR
	sts	sd_rsp+3, r24
	rcall	SPI_WRITE_DUMMY
	in	r24, SPDR
	sts	sd_rsp+4, r24
	rcall	SPI_CLEAR
	lds	r24, sd_status
	ori	r24, sd__ver2_bm	; Version 2 Card
	sts	sd_status, r24		; Update SD-Card status
SD_CARD_IF_ILL:				; Version 1 Card
	clc
	ret

SD_CARD_IF_ERR:
	rcall	SPI_CLEAR
	sts	sd_status, zero		; Reset SD-Card status
	sec
	ret

;
;	4) Initialise Card		Monitor Command: J
;
;	ACMD41 is used to query the initialisation status of a SD-Card, non-SD-cards
;	will not respond to ACMD41
;
SD_CARD_INIT:
;
;	We need to query the initialisation status until we get a ready status.
;	The ready status is bit0 of the response. If it is set then the SD-card
;	is idle and when cleared it is ready. All other status bits must be
;	cleared as well, else we have an error.
;
	clr	r18			; Number of Initialise Commands sent
SD_CARD_ACMD41_L1:
	rcall	SD_CARD_CMD55		; Send Prefix to prepare for ACMDs
	ldi	r24, 0x69
	sts	sd_cmd+0, r24
;
;	Cards that support HCS(High Capacity Support) can be switched to HCS
;	mode then the argument in read and write are no longer byte but block
;	offsets. Note for SDHC and SDXC cards HCS must be set in ACMD41 else
;	they will never switch from idle to ready.
;
	sts	sd_cmd+1, zero
	lds	r24, sd_status		; Response from CMD8
	andi	r24, sd__ver2_bm	; Version 2 Card ?
	breq	SD_CARD_ACMD41_V1	; Version 1 Cards have no HCS support
	ldi	r24, sd__ccs_bm		; Set bit to activate HCS support if available
SD_CARD_ACMD41_V1:
	sts	sd_cmd+1, r24
	sts	sd_cmd+2, zero
	sts	sd_cmd+3, zero
	sts	sd_cmd+4, zero
	ldi	r24, 1
	sts	sd_cmd+5, r24
	rcall	SD_SEND_CMD
	rcall	SPI_GET_RESPONSE	; R1 response
	sts	sd_rsp+0, r24
	cpi	r24, 0x00		; Initialisation done? I.e. no longer in idle mode
	breq	SD_CARD_ACMD41_DONE
	cpi	r24, 0x01		; The only other answer we accpet is Idle State
	brne	SD_CARD_ACMD41_E1	; Error

	rcall	SPI_CLEAR		; Still in Idle State
	inc	r18			; 
	brne	SD_CARD_ACMD41_L1
;
SD_CARD_ACMD41_E1:
	rcall	SPI_CLEAR
	lds	r24, sd_rsp+0
	sec
	ret
;
SD_CARD_ACMD41_DONE:
	lds	r24, sd_status		; Update SD-Card Satus
	ori	r24, sd__init_bm
	sts	sd_status, r24
	rcall	SPI_CLEAR
	mov	r24, r18		; Return retries
	clc
	ret
;
;	3) Read Operation Condition Register Monitor Command: K
;
;	CMD58, this is only required if the result sent in the reponse to CMD8 is
;	not acceptable. Note Version 1 SD Cards do respond to CMD8 with invalid 
;	command and also don't understand CMD58
;	
SD_CARD_READOCR:
	lds	r24, sd_status
	sbrs 	r24, sd__ver2_bp
	rjmp	SD_CARD_READOCR_V1	; 
	ldi	r24, 0x7A		; 0x40 + 58.
	sts	sd_cmd+0, r24
	sts	sd_cmd+1, zero
	sts	sd_cmd+2, zero
	sts	sd_cmd+3, zero
	sts	sd_cmd+4, zero
	ldi	r24, 1
	sts	sd_cmd+5, r24
	rcall	SD_SEND_CMD
	rcall	SPI_GET_RESPONSE
	sts	sd_rsp+0, r24
	cpi	r24, 0xff
	breq	SD_CARD_READOCRERR
	rcall	SPI_WRITE_DUMMY
	in	r24, SPDR
	sts	sd_rsp+1, r24
	sbrs	r24, sd__ccs_bp
	rjmp	SD_CARD_READOCR_NOHCS	; Card has no card capacity support
	lds	r24, sd_status
	ori	r24, sd__ccs_bm
	sts	sd_status, r24		; Set bit in status
SD_CARD_READOCR_NOHCS:
	rcall	SPI_WRITE_DUMMY
	in	r24, SPDR
	sts	sd_rsp+2, r24
	rcall	SPI_WRITE_DUMMY
	in	r24, SPDR
	sts	sd_rsp+3, r24
	rcall	SPI_WRITE_DUMMY
	in	r24, SPDR
	sts	sd_rsp+4, r24
	rcall	SPI_WRITE_DUMMY
	in	r24, SPDR
	sts	sd_rsp+5, r24
	rcall	SPI_CLEAR
SD_CARD_READOCR_V1:
	clc
	ret
SD_CARD_READOCRERR:
	rcall	SPI_CLEAR
	sec
	ret
;
;	5)	Set Block Length	Monitor Command: L
;
;	Set Block Length to 512bytes for standard SD-Cards
;
SD_CARD_BLKLEN:				; Set Block Length
	ldi	r24, 0x50		; Set Block Length Command
	sts	sd_cmd+0, r24
	sts	sd_cmd+1, zero
	sts	sd_cmd+2, zero
	ldi	r24, 0x02
	sts	sd_cmd+3, r24
	sts	sd_cmd+4, zero
	ldi	r24, 1
	sts	sd_cmd+5, r24
	rcall	SD_SEND_CMD
	rcall	SPI_GET_RESPONSE
	push	r24
	rcall	SPI_CLEAR
	pop	r24
	tst	r24
	breq	SD_CARD_BLKLEN_SUCC
	sec
	ret
SD_CARD_BLKLEN_SUCC:
	clc
	ret

.include "SD_CARD_RD_WR.asm"
