;--------------------------------------------------------------------------
;
;	Logging - Buffers, Definitions, Macro's
;
;	For debugging purposes a circular logging buffer is implemented. Each
;	log entry occupies 8-bytes. 
;
;	The logging buffer starts with an optional non-circular part where we 
;	will store the first log entries. Then follows a circular buffer which
;	stores the last few entries. The size of the non-circular buffer is   
;	defined by log_begin. The total logging buffer is defined by          
;	log_buffersize. Both, log_begin and log_buffersize must be a multiple 
;	of 256. The logging buffer must end at a boundary which is power of   
;	two so that an overflow of the log_pointer can be detected by a bit   
;	set in the log_pointer. E.g. if the buffer goes up to 0x3FF8 (each    
;	entry starts at a 8-byte boundary) and we add 8 to get the next       
;	log_pointer we just need to check bit 14 of the pointer, which is bit 
;	6 of the high byte. If this bit is set we know that we have reached   
;	the end of the circular buffer and need to reset the log_pointer to   
;	the start of the circular buffer. This is simply done by subtracting  
;	the high-byte of the circular buffer size, log_size, from log_pointer.           
;
.equ	log_buffer	= 0x3000	; The logging buffer start
.equ	log_buffersize	= 0x1000	; 4096 bytes of logging
.equ	log_begin	= 0x400		; Non-circular size must be multiple of 256
.equ	log_size	= log_buffersize - log_begin
.equ	log_overflow	= 6		; the 6th bit of the high byte address

.if (log_begin & 0xFF >0)
	.error " Invalid non-circular buffer size"
.endif
.if (log_buffersize & 0xFF >0)
	.error " Invalid total logging buffer size"
.endif
.if ((log_buffer & (log_buffersize - 1)) != 0)
	.error " Invalid logging buffer definitions"
.endif
;
;	Logging bits in GPIOR0
;
.equ	log__rl0	= 0		; Log IO to unit 0
.equ	log__rl1	= 1
.equ	log__rl2	= 2
.equ	log__rl3	= 3
.equ	log__units	= 0x0f		; All Units
.equ	log__rom	= 4		; Log ROM
.equ	log__iack	= 5		; Log Interrupt Ack (INTI)
.equ	log__pbn	= 6		; Log Physical Block Number
.equ	log__reg	= 7		; Log Device Register IO
;
;	Logging Message Types for non-function messages
;
#define	log_csrread	0		// put constant register zero
#define	log_csrwrite	1		// put constant register one
#define	log_pbn		2		// to be used with ldi
#define	log_iack	3
#define	log_romread	4
#define	log_romwrite	5
#define	log_autoboot	6
#define	log_init	7
#define	log_trace	8
#define	log_undef	9
;
;	The logging buffer must be naturally alligned and the
;	size must be a power of 2. With this we can wrap around
;	the address by using 'and' and 'or' with log_size and
;	log_buffer. 
;	However we can also use the following trick to get a
;	lollipop shaped logging buffer to keep the oldest
;	entries. We first check if new log_pointer points
;	past the log_buffer by checking the overflow bit.
;	
;	E.g. if log_buffer is 0x3000 and log_size is 0x0FFF
;	then the log_pointer must not be 0x4000 or higher, 
;	that is if bit 14 is set we have an overflow. Bit 14
;	is bit 6 of the high byte. We then just subtract
;	the log_size. By adding less than log_size we can
;	make sure that the first entries are never overwritten.
;
	.macro	lognxt
	adiw	@1:@0, 8		;  2
	sbrc	@1, log_overflow	;  2 sbrc and subi together always
	subi	@1, high(log_size)	;    require 2 cycles
	sts	log_pointer+0, @0	;  2
	sts	log_pointer+1, @1	;  2
	.endmacro
	
recordstart	msg
record		msg, type, 1
record		msg, ext, 1
record		msg, addr, 2
record		msg, data, 2
record		msg, wcount, 2
recordend	msg, size
;
;	Temporary logging message buffers and flags used by the RLV12
;	module
;
log_message:	.byte	8		; Logging message
log_pbnmsg:	.byte	8		; PBN Logging
log_pbnflag:	.byte	1		; Set if PBN is active
